from __future__ import annotations

from typing import List, NamedTuple, Sequence
from functools import partial
from scipy import stats
import numpy as np
import pandas as pd
from magicgui.widgets import (
    Container,
    ListEdit,
    PushButton,
    ComboBox,
    Label,
)

from tabulous._magicgui import find_current_table, SelectionWidget, ToggleSwitches
from tabulous._selection_op import SelectionOperator
from tabulous.exceptions import UnreachableError
from tabulous.widgets import Table


def result_table(result: NamedTuple) -> Table:
    """Create a Table widget from a named tuple"""
    index = []
    data = []
    for k, v in result._asdict().items():
        index.append(k)
        data.append(v)
    df = pd.DataFrame(data, index=index, columns=[type(result).__name__])
    return Table(df, editable=False)


combobox_alternative = partial(
    ComboBox,
    choices=["two-sided", "greater", "less"],
    value="two-sided",
    name="alternative",
    tooltip="Alternative hypothesis",
)

combobox_nan_policy = partial(
    ComboBox,
    choices=["propagate", "omit", "raise"],
    value="omit",
    name="nan_values",
    tooltip="How to handle NaN values",
)

# ########################################################
#   Data container
# ########################################################


class DataContainer(Container):
    def extract_dataset(self, df: pd.DataFrame) -> dict[str, np.ndarray]:
        raise NotImplementedError


class MeltedDataContainer(DataContainer):
    """Data container for melted data, a column for labels and one for data."""

    def __init__(self, **kwargs):
        self._label_range_wdt = SelectionWidget(format="iloc", name="labels")
        self._data_range_wdt = SelectionWidget(format="iloc", name="data")
        super().__init__(
            widgets=[self._label_range_wdt, self._data_range_wdt], **kwargs
        )
        self.margins = (0, 0, 0, 0)

    def extract_dataset(self, df: pd.DataFrame) -> dict[str, np.ndarray]:
        label = self._label_range_wdt.value.operate(df)
        data = self._data_range_wdt.value.operate(df)

        if label.shape[1] != 1:
            raise ValueError("Label must be a single column")
        if data.shape[1] != 1:
            raise ValueError("Data must be a single column")
        if label.size != data.size:
            raise ValueError("Label and data must have the same size")

        label = np.asarray(label).ravel()
        data = np.asarray(data).ravel()
        unique_labels = np.unique(label)
        return {_l: data[label == _l] for _l in unique_labels}


class UnstructuredDataContainer(DataContainer):
    """Data container for unstructured data, each column for each data."""

    def __init__(self, **kwargs):
        self._data_range_list = ListEdit(
            value=[None, None],
            annotation=List[SelectionOperator],
            name="data",
            layout="vertical",
            options=dict(format="iloc"),
        )
        super().__init__(widgets=[self._data_range_list], **kwargs)
        self.margins = (0, 0, 0, 0)

    def extract_dataset(self, df: pd.DataFrame) -> dict[str, np.ndarray]:
        data_dict = {}
        for op in self._data_range_list.value:
            op: SelectionOperator
            data = op.operate(df)
            if data.shape[1] != 1:
                raise ValueError("Data must be a single column")
            data_dict[data.columns[0]] = np.asarray(data).ravel()
        return data_dict


# ########################################################
#   Parameters for each test
# ########################################################


class TestRunner(Container):
    def run_test(self, data: Sequence[np.ndarray]) -> NamedTuple:
        raise NotImplementedError


class TtestRunner(TestRunner):
    """Parameters for t-test"""

    def __init__(self, **kwargs):
        self._test_type = ComboBox(
            choices=["Student's", "Welch's", "Related"], name="type"
        )
        self._alternative = combobox_alternative()
        self._nan_policy = combobox_nan_policy()
        super().__init__(
            widgets=[self._test_type, self._alternative, self._nan_policy],
            **kwargs,
        )
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        if len(data) != 2:
            raise ValueError("t-test only supports two data sets")
        kwargs = dict(
            alternative=self._alternative.value,
            nan_policy=self._nan_policy.value,
        )
        if self._test_type.value == "Student's":
            return stats.ttest_ind(data[0], data[1], **kwargs)
        elif self._test_type.value == "Welch's":
            return stats.ttest_ind(data[0], data[1], equal_var=False, **kwargs)
        elif self._test_type.value == "Related":
            return stats.ttest_rel(data[0], data[1], **kwargs)
        else:
            raise UnreachableError(self._test_type)


class MannwhitneyuRunner(TestRunner):
    """Parameters for Mann-Whitney U-test"""

    def __init__(self, **kwargs):
        self._alternative = combobox_alternative()
        self._nan_policy = combobox_nan_policy()
        super().__init__(widgets=[self._alternative, self._nan_policy], **kwargs)
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        if len(data) != 2:
            raise ValueError("Mann-Whitney U-test only supports two data sets")
        return stats.mannwhitneyu(
            data[0],
            data[1],
            alternative=self._alternative.value,
            nan_policy=self._nan_policy.value,
        )


class KruskalRunner(TestRunner):
    """Parameters for Kruskal-Wallis H test"""

    def __init__(self, **kwargs):
        self._nan_policy = combobox_nan_policy()
        super().__init__(widgets=[self._nan_policy], **kwargs)
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        return stats.kruskal(*data, nan_policy=self._nan_policy.value)


class WilcoxonRunner(TestRunner):
    """Parameters for Wilcoxon signed-rank test"""

    def __init__(self, **kwargs):
        self._alternative = combobox_alternative()
        self._nan_policy = combobox_nan_policy()
        super().__init__(widgets=[self._alternative, self._nan_policy], **kwargs)
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        if len(data) != 2:
            raise ValueError("Wilcoxon signed-rank test only supports two data sets")
        return stats.wilcoxon(
            data[0],
            data[1],
            alternative=self._alternative.value,
            nan_policy=self._nan_policy.value,
        )


class FriedmanRunner(TestRunner):
    """Parameters for Friedman test"""

    def __init__(self, **kwargs):
        super().__init__(widgets=[], **kwargs)
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        return stats.friedmanchisquare(*data)


class AnovaRunner(TestRunner):
    """Parameters for ANOVA"""

    def __init__(self, **kwargs):
        super().__init__(widgets=[], **kwargs)
        self.margins = (0, 0, 0, 0)

    def run_test(self, data: Sequence[np.ndarray]):
        return stats.f_oneway(*data)


# ########################################################
#   Main widget
# ########################################################


class DataTypes:
    """Choices for data types"""

    SEPARATE = "Separate data"
    LABELED = "Labeled data"


class TestTypes:
    """Choices for test types"""

    T = "Student's t-test"
    U = "Mann-Whitney U-test"
    H = "Kruskal-Wallis H-test"
    Wilcoxon = "Wilcoxon signed-rank test"
    Friedman = "Friedman test"
    ANOVA = "one-way ANOVA"

    _all = [T, U, H, Wilcoxon, Friedman, ANOVA]


class StatsTestDialog(Container):
    def __init__(self, **kwargs):
        self._data_types = ToggleSwitches(
            choices=[DataTypes.SEPARATE, DataTypes.LABELED], value=DataTypes.SEPARATE
        )
        self._data_containers: dict[str, DataContainer] = {
            DataTypes.SEPARATE: UnstructuredDataContainer(),
            DataTypes.LABELED: MeltedDataContainer(),
        }

        self._test_types = ComboBox(
            choices=TestTypes._all,
            value=TestTypes.T,
        )
        self._test_runners: dict[str, TestRunner] = {
            TestTypes.T: TtestRunner(),
            TestTypes.U: MannwhitneyuRunner(),
            TestTypes.H: KruskalRunner(),
            TestTypes.Wilcoxon: WilcoxonRunner(),
            TestTypes.Friedman: FriedmanRunner(),
            TestTypes.ANOVA: AnovaRunner(),
        }

        self._call_button = PushButton(text="Run")

        super().__init__(
            widgets=[
                Label(value="Input datasets"),
                self._data_types,
                *self._data_containers.values(),
                Label(value="Test method and parameters"),
                self._test_types,
                *self._test_runners.values(),
                self._call_button,
            ],
            **kwargs,
        )

        self._call_button.changed.connect(self._on_call)
        self._data_types.changed.connect(self._on_data_type_changed)
        self._test_types.changed.connect(self._on_test_type_changed)
        self._on_data_type_changed(DataTypes.SEPARATE)
        self._on_test_type_changed(TestTypes.T)

    def _on_call(self):
        table = find_current_table(self)
        if table is None:
            raise ValueError("No table found")
        df = table.data
        widget = self._data_containers[self._data_types.value]

        data_dict = widget.extract_dataset(df)
        self.close()

        data_input = list(data_dict.values())
        results = self._test_runners[self._test_types.value].run_test(data_input)
        table.add_side_widget(result_table(results), name="Test result")

    def _on_data_type_changed(self, v):
        for k, w in self._data_containers.items():
            w.visible = k == v

    def _on_test_type_changed(self, v):
        for k, w in self._test_runners.items():
            w.visible = k == v



from __future__ import annotations

from typing import TYPE_CHECKING
from enum import Enum
import numpy as np
from magicgui.widgets import (
    LineEdit,
    Container,
    LiteralEvalLineEdit,
    PushButton,
    Widget,
)
from tabulous._magicgui import SelectionWidget
from tabulous.exceptions import UnreachableError
from tabulous._magicgui import ToggleSwitches

if TYPE_CHECKING:
    import pandas as pd


class _RandomGenerator(Container):
    def __init__(self, **kwargs):
        widgets = self.prepare_widgets()
        super().__init__(widgets=widgets, **kwargs)

    def prepare_widgets(self) -> list[Widget]:
        return []

    def generate(self, shape: tuple[int, int]) -> np.ndarray:
        raise NotImplementedError


class UniformRandomGenerator(_RandomGenerator):
    def prepare_widgets(self) -> list[Widget]:
        return [
            LineEdit(value="0", label="minimum"),
            LineEdit(value="1", label="maximum"),
        ]

    def generate(self, shape: tuple[int, int]) -> np.ndarray:
        val = np.random.random(shape)
        mn = float(self[0].value)
        mx = float(self[1].value)
        return mn + (mx - mn) * val


class ChoiceRandomGenerator(_RandomGenerator):
    def prepare_widgets(self) -> list[Widget]:
        return [
            LiteralEvalLineEdit(
                value="[0, 1, 2]",
                label="choices",
                tooltip="Iterable object that provide choices",
            ),
        ]

    def generate(self, shape: tuple[int, int]) -> np.ndarray:
        choices = self[0].value
        return np.random.choice(choices, shape)


class NormalRandomGenerator(_RandomGenerator):
    def prepare_widgets(self) -> list[Widget]:
        return [
            LineEdit(value="0", label="mean"),
            LineEdit(value="1", label="sigma"),
        ]

    def generate(self, shape: tuple[int, int]) -> np.ndarray:
        return np.random.normal(
            float(self[0].value),
            float(self[1].value),
            size=shape,
        )


class PoissonRandomGenerator(_RandomGenerator):
    def prepare_widgets(self) -> list[Widget]:
        return [
            LineEdit(value="1.0", label="mean"),
        ]

    def generate(self, shape: tuple[int, int]) -> np.ndarray:
        return np.random.poisson(
            float(self[0].value),
            size=shape,
        )


class Generator(Enum):
    uniform = "uniform"
    normal = "normal"
    poisson = "poisson"
    choices = "choices"


class RandomGeneratorDialog(Container):
    def __init__(self):
        self._selection_wdt = SelectionWidget(
            format="iloc", allow_out_of_bounds=True, label="Selection"
        )
        self._radio_buttons = ToggleSwitches(
            choices=Generator, value=Generator.uniform, label="Distribution"
        )
        self._uniform_wdt = UniformRandomGenerator(label="parameters")
        self._normal_wdt = NormalRandomGenerator(label="parameters")
        self._poisson_wdt = PoissonRandomGenerator(label="parameters")
        self._choices_wdt = ChoiceRandomGenerator(label="parameters")
        self._call_button = PushButton(text="Generate data")

        super().__init__(
            widgets=[
                self._selection_wdt,
                self._radio_buttons,
                self._uniform_wdt,
                self._normal_wdt,
                self._poisson_wdt,
                self._choices_wdt,
                self._call_button,
            ]
        )

        self._radio_buttons.changed.connect(self._on_choice_changed)
        self._radio_buttons.changed.emit(Generator.uniform)
        self.called = self._call_button.changed
        self.called.connect(self.close)

    def _on_choice_changed(self, choice: Generator):
        self._uniform_wdt.visible = choice == Generator.uniform
        self._normal_wdt.visible = choice == Generator.normal
        self._poisson_wdt.visible = choice == Generator.poisson
        self._choices_wdt.visible = choice == Generator.choices

    def _current_widget(self) -> _RandomGenerator:
        if self._radio_buttons.value is Generator.uniform:
            return self._uniform_wdt
        elif self._radio_buttons.value is Generator.normal:
            return self._normal_wdt
        elif self._radio_buttons.value is Generator.poisson:
            return self._poisson_wdt
        elif self._radio_buttons.value is Generator.choices:
            return self._choices_wdt
        else:
            raise UnreachableError(self._radio_buttons.value)

    def get_value(self, df: pd.DataFrame) -> tuple[slice, slice, np.ndarray]:
        selop = self._selection_wdt.value
        shape = selop.shape(df)
        random_data = self._current_widget().generate(shape)
        rslice, cslice = selop.as_iloc_slices(df)
        return rslice, cslice, random_data



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, overload, Any, Iterator, List

import numpy as np
from psygnal import Signal, SignalGroup
from tabulous.exceptions import TableImmutableError
from tabulous.types import HeaderInfo
from ._base import Component, TableComponent
from tabulous.widgets._registry import SupportActionRegistration

if TYPE_CHECKING:
    import pandas as pd
    from tabulous.widgets import TableBase  # noqa: F401
    from tabulous._qt._table._base._header_view import QDataFrameHeaderView


class HeaderEvents(SignalGroup):
    selected = Signal(List[slice])
    renamed = Signal(HeaderInfo)


class HeaderSectionSpan(Component["_HeaderInterface"]):
    def __getitem__(self, index: int) -> int:
        header = self.parent._get_header()
        return header.sectionSize(index)

    def __setitem__(
        self,
        index: int | slice | list[int],
        span: int | Sequence[int],
    ) -> None:
        header = self.parent._get_header()
        if isinstance(index, (slice, list)):
            if isinstance(index, slice):
                index = list(range(*index.indices(header.count())))
            if isinstance(span, Sequence):
                # set span for each section
                if len(span) != len(index):
                    raise ValueError("Size mismatch between destination and spans.")
                [header.resizeSection(idx, sp) for idx, sp in zip(index, span)]
            else:
                [header.resizeSection(idx, span) for idx in index]
        else:
            header.resizeSection(index, span)

    def __iter__(self) -> Iterator[int]:
        return iter(int(a) for a in self.parent._get_header()._section_sizes)

    def __eq__(self, other: Any) -> bool:
        return np.all(
            self.parent._get_header()._section_sizes.astype(np.uint16) == other
        )

    def resize_to_content(self):
        from qtpy.QtWidgets import QHeaderView

        header = self.parent._get_header()
        header.resizeSections(QHeaderView.ResizeMode.ResizeToContents)
        return None

    def __repr__(self) -> str:
        sizes = self.parent._get_header()._section_sizes
        return f"{type(self).__name__}({sizes.astype(np.uint16).tolist()!r})"


class _HeaderInterface(TableComponent, SupportActionRegistration["TableBase", int]):
    """
    Interface to the table {index/columns} header.

    Similar to pandas.DataFrame.{index/columns}, you can get/set labels.

    >>> print(table.{index/columns}[0])
    >>> table.{index/columns}[0] = 'new_label'

    Use :meth:`selected` to get/set selected ranges.

    >>> table.{index/columns}.selected
    >>> table.{index/columns}.selected = [0, 1, 2]
    >>> table.{index/columns}.selected = [slice(0, 3), slice(6, 8)]

    Use :meth:`insert` / :meth:`remove` to run insert/remove command in
    spreadsheet.

    >>> table.{index/columns}.insert(at=0, count=2)
    >>> table.{index/columns}.remove(at=0, count=2)

    Use ``register`` to register a contextmenu function.

    >>> @table.{index/columns}.register("My Action")
    >>> def my_action(table, index):
    ...     # do something

    Many other pandas.Index methods are also available.

    >>> table.{index/columns}.str.contains('foo')
    >>> table.{index/columns}.isin(['foo', 'bar'])
    >>> table.{index/columns}.get_loc('foo')
    """

    _AXIS_NUMBER: int

    def __init__(self, parent: TableBase = TableComponent._no_ref):
        super().__init__(parent)
        self.events = HeaderEvents()

    def _get_axis(self) -> pd.Index:
        raise NotImplementedError()

    def _set_value(self, idx: int, val: Any):
        raise NotImplementedError()

    def _get_header(self) -> QDataFrameHeaderView:
        raise NotImplementedError()

    @property
    def _qcontextmenu(self):
        """The QContextMenu widget."""
        return self._get_header()._qt_context_menu

    def __repr__(self) -> str:
        return f"<{type(self).__name__}({self._get_axis()!r}) of {self.parent!r}>"

    # fmt: off
    @overload
    def __getitem__(self, key: int) -> str: ...
    @overload
    def __getitem__(self, key: slice | list[int] | np.ndarray) -> Sequence[str]: ...
    # fmt: on

    def __getitem__(self, key):
        return self._get_axis()[key]

    def __setitem__(self, key: int | slice, value: Any):
        table = self.parent
        if not table.editable:
            raise TableImmutableError("Table is not editable.")

        if isinstance(key, slice):
            start, stop, step = key.indices(len(self._get_axis()))
            for i, idx in enumerate(range(start, stop, step)):
                self._set_value(idx, value[i])
        else:
            idx = key.__index__()
            self._set_value(idx, value)
        return None

    def __len__(self) -> int:
        return len(self._get_axis())

    def __iter__(self) -> Iterator[str]:
        return iter(self._get_axis())

    def __contains__(self, key: str) -> bool:
        return key in self._get_axis()

    def __eq__(self, other: Sequence[str]) -> np.ndarray:
        return self._get_axis() == other

    @property
    def str(self):
        """The string accessor to the header."""
        return self._get_axis().str

    @property
    def size(self) -> int:
        """Size of the axis."""
        return self._get_axis().size

    def get_loc(self, key: str) -> int:
        """Get the location of a label."""
        return self._get_axis().get_loc(key)

    def isin(self, values) -> np.ndarray:
        """Return a boolean array of whether each is found in the passed values."""
        return self._get_axis().isin(values)

    def coerce_name(self, name: str, start: int | None = None) -> str:
        """Coerce a name to avoid name collision."""
        index = self._get_axis()
        i = 0
        if start is not None:
            name = f"{name}_{start}"
            i = start + 1
        while name in index:
            name = f"{name}_{i}"
            i += 1
        return name

    def _get_qregistry(self):
        return self._get_header()

    @property
    def selected(self) -> list[slice]:
        """Return the selected ranges."""
        smodel = self.parent._qwidget._qtable_view._selection_model
        out = [
            smodel.ranges[i][self._AXIS_NUMBER] for i in smodel._col_selection_indices
        ]
        return out

    @property
    def span(self) -> HeaderSectionSpan:
        """Sub-field to interact with section spans."""
        return HeaderSectionSpan(self)


class VerticalHeaderInterface(_HeaderInterface):
    __doc__ = _HeaderInterface.__doc__.replace("{index/columns}", "index")
    _AXIS_NUMBER = 0

    def _get_axis(self) -> pd.Index:
        return self.parent._qwidget.model().df.index

    def _set_value(self, idx: int, val: Any):
        return self.parent._qwidget.setVerticalHeaderValue(idx, val)

    def _get_header(self) -> QDataFrameHeaderView:
        return self.parent._qwidget._qtable_view.verticalHeader()

    def insert(self, at: int, count: int = 1):
        """Insert `count` rows at the given position."""
        sheet = self._assert_spreadsheet()
        with sheet._qwidget._anim_row.using_animation(False):
            sheet._qwidget.insertRows(at, count)

    def remove(self, at: int, count: int = 1):
        """Remove `count` rows at the given position."""
        sheet = self._assert_spreadsheet()
        with sheet._qwidget._anim_row.using_animation(False):
            sheet._qwidget.removeRows(at, count)

    @_HeaderInterface.selected.setter
    def selected(self, slices: int | slice | list[int | slice]):
        """Set the selected ranges."""
        if isinstance(slices, (int, slice)):
            slices = [slices]
        smodel = self.parent._qwidget._qtable_view._selection_model
        smodel.clear()
        csl = slice(0, smodel._col_count_getter())
        smodel.extend(
            ((_as_slice(sl), csl) for sl in slices),
            row=True,
        )
        self.parent.refresh()


class HorizontalHeaderInterface(_HeaderInterface):
    __doc__ = _HeaderInterface.__doc__.replace("{index/columns}", "columns")
    _AXIS_NUMBER = 1

    def _get_axis(self) -> pd.Index:
        return self.parent._qwidget.model().df.columns

    def _set_value(self, idx: int, val: Any):
        return self.parent._qwidget.setHorizontalHeaderValue(idx, val)

    def _get_header(self) -> QDataFrameHeaderView:
        return self.parent._qwidget._qtable_view.horizontalHeader()

    def insert(self, at: int, count: int = 1):
        """Insert `count` columns at the given position."""
        sheet = self._assert_spreadsheet()
        with sheet._qwidget._anim_col.using_animation(False):
            sheet._qwidget.insertColumns(at, count)

    def remove(self, at: int, count: int = 1):
        """Remove `count` columns at the given position."""
        sheet = self._assert_spreadsheet()
        with sheet._qwidget._anim_col.using_animation(False):
            sheet._qwidget.removeColumns(at, count)

    @_HeaderInterface.selected.setter
    def selected(self, slices: int | slice | list[int | slice]):
        """Set the selected ranges."""
        if isinstance(slices, (int, slice)):
            slices = [slices]
        smodel = self.parent._qwidget._qtable_view._selection_model
        smodel.clear()
        rsl = slice(0, smodel._row_count_getter())
        smodel.extend(
            ((rsl, _as_slice(sl)) for sl in slices),
            column=True,
        )
        self.parent.refresh()


def _as_slice(x: int | slice) -> slice:
    if isinstance(x, slice):
        return x
    else:
        return slice(x, x + 1)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence, TypeVar, Union, overload, Callable

import numpy as np
from tabulous.exceptions import TableImmutableError

from tabulous.types import ColorMapping
from ._base import TableComponent
from ._column_setting import _Void

if TYPE_CHECKING:
    from numpy.typing import NDArray
    import pandas as pd
    from tabulous.widgets import TableBase, Table, SpreadSheet
    from ._column_setting import (
        _Interpolatable,
        _ColormapInterface,
        _DictPropertyInterface,
    )

    _RowLocIdx = Union[int, str, slice, list[str], list[int]]
    _ColumnLocIdx = Union[str, slice, list[str]]
    _ILocIdx = Union[slice, list[int], NDArray[np.integer]]


class TableLocIndexer(TableComponent):
    """
    loc indexer for Table widget.

    >>> sheet = viewer.add_spreadsheet(np.zeros((10, 5)))
    >>> sheet.loc[:, "B":"C"].data
    """

    # fmt: off
    @overload
    def __getitem__(self, key: _RowLocIdx | tuple[_RowLocIdx, _ColumnLocIdx]) -> TableSubset: ...  # noqa: E501
    @overload
    def __getitem__(self, key: tuple[_RowLocIdx, str]) -> TableSeries: ...
    # fmt: on

    def __getitem__(self, key):
        table = self.parent
        if not isinstance(key, tuple):
            ckey = list(self.parent.columns)
            rkey = key
        else:
            rkey, ckey = key

        if isinstance(rkey, (int, str)):
            idx = table.index.get_loc(rkey)
            rsl = slice(idx, idx + 1)
        elif isinstance(rkey, slice):
            if rkey.start is None:
                start = None
            else:
                start = table.index.get_loc(rkey.start)
            if rkey.stop is None:
                stop = None
            else:
                stop = table.index.get_loc(rkey.stop) + 1
            rsl = slice(start, stop)
        elif isinstance(rkey, Sequence):
            rsl = [table.index.get_loc(lbl) for lbl in rkey]
        else:
            raise TypeError(f"Cannot loc-slice by {type(ckey)}")

        if isinstance(ckey, str):
            if ckey not in table.columns:
                raise KeyError(ckey)
            return TableSeries(table, rsl, ckey)
        elif isinstance(ckey, slice):
            if ckey.start is None:
                start = None
            else:
                start = table.columns.get_loc(ckey.start)
            if ckey.stop is None:
                stop = None
            else:
                stop = table.columns.get_loc(ckey.stop) + 1
            csl = slice(start, stop)
            columns = table.columns[csl]
            return TableSubset(table, rsl, columns)
        elif isinstance(ckey, Sequence):
            for ck in ckey:
                if ck not in table.columns:
                    raise KeyError(ckey)
            return TableSubset(table, rsl, ckey)
        else:
            raise TypeError(f"Cannot loc-slice by {type(ckey)}")


class TableILocIndexer(TableComponent):
    """
    iloc indexer for Table widget.

    >>> sheet = viewer.add_spreadsheet(np.zeros((10, 5)))
    >>> sheet.iloc[:, 2:5].data
    """

    # fmt: off
    @overload
    def __getitem__(self, key: _ILocIdx | tuple[_ILocIdx, _ILocIdx]) -> TableSubset: ...
    @overload
    def __getitem__(self, key: tuple[_ILocIdx, int]) -> TableSeries: ...
    # fmt: on

    def __getitem__(self, key):
        table = self.parent
        if not isinstance(key, tuple):
            ckey = slice(None)
            rkey = key
        else:
            rkey, ckey = key

        if isinstance(ckey, int):
            return TableSeries(table, rkey, table.columns[ckey])
        elif isinstance(ckey, (slice, Sequence)):
            return TableSubset(table, rkey, list(table.columns[ckey]))
        else:
            raise TypeError(f"Cannot iloc-slice by {type(ckey)}")


class DataProperty:
    """Internal subset data of the table."""

    def __get__(self, instance: TableSubset, owner) -> pd.DataFrame:
        if instance is None:
            raise AttributeError("Cannot access property without instance.")
        table = instance.parent
        return table.native._get_sub_frame(instance._columns).iloc[instance._row_slice]

    def __set__(self, instance: TableSubset, val: Any):
        if instance is None:
            raise AttributeError("Cannot access property without instance.")
        import pandas as pd

        table = instance.parent
        if not table.mutable:
            raise TableImmutableError(f"Cannot set data on {type(table)}")
        table: Table | SpreadSheet
        df = table.native._get_sub_frame(instance._columns)
        to_assign = {}
        if not isinstance(val, pd.DataFrame):
            val = pd.DataFrame(val, columns=df.columns)
        for k, v in df.items():
            v.values[instance._row_slice] = val[k]
            to_assign[k] = v
        table.assign(to_assign)
        return None


class TableSubset(TableComponent):
    """Object representing a subset of a Table widget."""

    def __init__(
        self, parent: TableBase, row_slice: slice | list[int], columns: list[str]
    ):
        super().__init__(parent)
        self._row_slice = row_slice
        self._columns = columns

    data = DataProperty()

    def __repr__(self) -> str:
        return f"<TableSubset of {self.parent!r}> with data:\n{self.data!r}"

    def _ipython_key_completions_(self):
        return self._columns


class TableSeries(TableComponent):
    """Object representing a column series of a Table widget."""

    def __init__(self, parent: TableBase, row_slice: slice | list[int], column: str):
        super().__init__(parent)
        self._row_slice = row_slice
        self._column = column

    def __repr__(self) -> str:
        return f"<TableSeries of {self.parent!r}> with data:\n{self.data!r}"

    @property
    def data(self) -> pd.Series:
        """Data of the column as a pandas Series."""
        table = self.parent
        return table.native._get_sub_frame(self._column).iloc[self._row_slice]

    @data.setter
    def data(self, val) -> None:
        table = self.parent
        if not table.mutable:
            raise TableImmutableError(f"Cannot set data on {type(table)}")
        table: Table | SpreadSheet
        ds = table.native._get_sub_frame(self._column)
        ds[self._row_slice] = val
        table.assign({ds.name: ds})
        return None

    @property
    def text_color(self):
        """Get the text colormap of the column."""
        self._assert_row_not_sliced()
        return PartialTextColormapInterface(self.parent, self._column)

    @text_color.setter
    def text_color(self, val) -> None:
        """Set the text colormap of the column."""
        self._assert_row_not_sliced()
        self.parent.text_color[self._column] = val
        return None

    @text_color.deleter
    def text_color(self) -> None:
        """Delete the text colormap of the column."""
        self._assert_row_not_sliced()
        del self.parent.text_color[self._column]
        return None

    @property
    def background_color(self):
        """Get the background colormap of the column."""
        self._assert_row_not_sliced()
        return PartialBackgroundColormapInterface(self.parent, self._column)

    @background_color.setter
    def background_color(self, val):
        """Set the background colormap of the column."""
        self._assert_row_not_sliced()
        self.parent.background_color[self._column] = val
        return None

    @background_color.deleter
    def background_color(self):
        """Delete the background colormap of the column."""
        self._assert_row_not_sliced()
        del self.parent.background_color[self._column]
        return None

    @property
    def formatter(self):
        self._assert_row_not_sliced()
        return PartialTextFormatterInterface(self.parent, self._column)

    @formatter.setter
    def formatter(self, val):
        self._assert_row_not_sliced()
        self.parent.formatter[self._column] = val
        return None

    @formatter.deleter
    def formatter(self):
        self._assert_row_not_sliced()
        del self.parent.formatter[self._column]
        return None

    @property
    def validator(self):
        self._assert_row_not_sliced()
        return PartialValidatorInterface(self.parent, self._column)

    @validator.setter
    def validator(self, val):
        self._assert_row_not_sliced()
        self.parent.validator[self._column] = val
        return None

    @validator.deleter
    def validator(self):
        self._assert_row_not_sliced()
        del self.parent.validator[self._column]
        return None

    def _assert_row_not_sliced(self):
        if self._row_slice == slice(None):
            return
        raise ValueError(f"{self!r} is sliced in row axis.")


_F = TypeVar("_F", bound=Callable)


class _PartialInterface(TableComponent):
    def _get_field(self) -> _DictPropertyInterface:
        raise NotImplementedError()

    def __init__(self, parent: TableBase, column: str):
        super().__init__(parent)
        self._column = column

    def __repr__(self) -> str:
        pclsname = type(self._get_field()).__name__
        item = self.item()
        if item is None:
            return f"{pclsname}<{self._column!r}>(Undefined)"
        else:
            return f"{pclsname}<{self._column!r}>({item})"

    def set(self, func: _F = _Void):
        return self._get_field().set(self._column, func)

    def reset(self):
        return self._get_field().reset(self._column)

    def item(self):
        return self._get_field().get(self._column, None)


class _PartialColormapInterface(_PartialInterface):
    def _get_field(self) -> _ColormapInterface:
        raise NotImplementedError()

    def invert(self):
        """Invert the colormap of the column."""
        return self._get_field().invert(self._column)

    def set_opacity(self, opacity: float):
        """Set the opacity of the colormap of the column."""
        return self._get_field().set_opacity(self._column, opacity)

    def adjust_brightness(self, factor: float):
        """Change the brightness of the colormap of the column."""
        return self._get_field().adjust_brightness(self._column, factor)

    def set(
        self,
        colormap: ColorMapping = _Void,
        *,
        interp_from: _Interpolatable | None = None,
        infer_parser: bool = True,
        opacity: float | None = None,
    ):
        return self._get_field().set(
            self._column,
            colormap,
            interp_from=interp_from,
            infer_parser=infer_parser,
            opacity=opacity,
        )


class PartialTextColormapInterface(_PartialColormapInterface):
    def _get_field(self):
        return self.parent.text_color


class PartialBackgroundColormapInterface(_PartialColormapInterface):
    def _get_field(self):
        return self.parent.background_color


class PartialTextFormatterInterface(_PartialInterface):
    def _get_field(self):
        return self.parent.formatter


class PartialValidatorInterface(_PartialInterface):
    def _get_field(self):
        return self.parent.validator



from __future__ import annotations

import argparse
from pathlib import Path


class TabulousArgs(argparse.Namespace):
    """Tabulous specific arguments"""

    profile: bool
    user_dir: bool
    debug: bool
    init_config: bool
    init_history: bool
    open_file: str | None


class TabulousParser(argparse.ArgumentParser):
    """Tabulous specific argument parser"""

    def __init__(self):
        from . import __version__

        super().__init__(description="Command line interface of tabulous.")
        self.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"tabulous version {__version__}",
        )
        self.add_argument("--profile", action="store_true")
        self.add_argument("--user-dir", action="store_true")
        self.add_argument("--debug", action="store_true")
        self.add_argument("--init-config", action="store_true")
        self.add_argument("--init-history", action="store_true")

    def parse_known_args(
        self, args=None, namespace=None
    ) -> tuple[TabulousArgs, list[str]]:
        args, unknown = super().parse_known_args(args, namespace)
        args = TabulousArgs(**vars(args))
        if not unknown:
            args.open_file = None
        elif len(unknown) == 1:
            path = unknown[0]
            if Path(path).suffix:
                args.open_file = path
            else:
                args.open_file = None
        else:
            raise argparse.ArgumentError("Too many arguments.")
        return args, unknown


def main():
    parser = TabulousParser()

    args, _ = parser.parse_known_args()

    if args.profile or args.user_dir:
        from ._utils import CONFIG_PATH

        return print(CONFIG_PATH.parent)

    if args.debug:
        import logging

        logger = logging.getLogger("tabulous")
        logging.basicConfig(format="%(levelname)s|| %(message)s")
        logger.setLevel(logging.DEBUG)

    if args.init_config:
        from ._utils import CONFIG_PATH, TabulousConfig

        CONFIG_PATH.unlink(missing_ok=True)
        TabulousConfig.from_toml(CONFIG_PATH).as_toml()
        return print(f"tabulous config file initialized at {str(CONFIG_PATH)}.")

    if args.init_history:
        from ._utils import TXT_PATH

        TXT_PATH.write_text("")

    from . import TableViewer

    viewer = TableViewer()

    if args.open_file:
        viewer.open(args.open_file)

    viewer.show()
    return


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any, Callable, Generator, overload, TypeVar
from typing_extensions import ParamSpec
import inspect
from functools import lru_cache
from superqt.utils import (
    thread_worker as _thread_worker,
    GeneratorWorker,
    FunctionWorker,
)

__all__ = ["thread_worker"]

_Y = TypeVar("_Y")
_S = TypeVar("_S")
_R = TypeVar("_R")
_P = ParamSpec("_P")


@overload
def thread_worker(
    function: Callable[_P, Generator[_Y, _S, _R]],
    *,
    desc: str | None = None,
    total: int = 0,
) -> Callable[_P, GeneratorWorker[_Y, _S, _R]]:
    ...


@overload
def thread_worker(
    function: Callable[_P, _R],
    *,
    desc: str | None = None,
    total: int = 0,
) -> Callable[_P, FunctionWorker[_R]]:
    ...


@overload
def thread_worker(
    function: None = None,
    *,
    desc: str | None = None,
    total: int = 0,
) -> Callable[
    [Callable[_P, _R]], Callable[_P, FunctionWorker[_R] | GeneratorWorker[Any, Any, _R]]
]:
    ...


def thread_worker(function=None, *, desc=None, total=0):
    """
    Convert the returned value of a function into a worker.

    >>> from tabulous.threading import thread_worker
    >>> @thread_worker
    >>> def func():
    ...     time.sleep(1)

    Parameters
    ----------
    function : callable
        Function to be called in another thread.
    desc : str, optional
        Label that will shown beside the progress indicator. The function name
        will be used if not provided.
    total : int, default is 0
        Total number of steps for the progress indicator.
    """

    def _inner(fn: Callable):
        return create_worker(fn, desc=desc, total=total)

    return _inner if function is None else _inner(function)


def create_worker(
    fn: Callable,
    *,
    desc: str | None = None,
    total: int = 0,
):
    worker_constructor = _thread_worker(fn)
    sig = inspect.signature(fn)

    def _create_worker(*args, **kwargs):
        from tabulous._qt._mainwindow import QMainWindow

        nonlocal desc, total

        bound = sig.bind_partial(*args, **kwargs)
        if desc is None:
            _desc = getattr(fn, "__name__", repr(fn))
        elif callable(desc):
            _desc = _call_with_filtered(desc, bound.arguments)
        else:
            _desc = desc
        if not isinstance(_desc, str):
            raise TypeError("`desc` did not return a str.")

        if callable(total):
            _total = _call_with_filtered(total, bound.arguments)
        else:
            _total = total
        if not isinstance(_total, int):
            raise TypeError("`total` did not return an int.")
        viewer = QMainWindow.currentViewer()
        worker = worker_constructor(*args, **kwargs)
        viewer.native._tablestack._info_stack.addWorker(
            worker, desc=_desc, total=_total
        )
        return worker

    return _create_worker


@lru_cache(maxsize=32)
def _make_filter(fn: Callable[..., _R]) -> Callable[[dict[str, Any]], dict[str, Any]]:
    sig = inspect.signature(fn)
    arg_names: list[str] = []
    for name, param in sig.parameters.items():
        if param.kind in (
            param.POSITIONAL_ONLY,
            param.POSITIONAL_OR_KEYWORD,
            param.KEYWORD_ONLY,
        ):
            arg_names.append(name)
        elif param.kind == param.VAR_POSITIONAL:
            raise NotImplementedError("Cannot use *args or **kwargs")
        elif param.kind == param.VAR_KEYWORD:
            raise NotImplementedError("Cannot use *args or **kwargs")

    def _filter(kwargs: dict):
        return {k: v for k, v in kwargs.items() if k in arg_names}

    return _filter


def _call_with_filtered(fn: Callable[..., _R], kwargs: dict[str, Any]) -> _R:
    filt = _make_filter(fn)
    return fn(**filt(kwargs))



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, SupportsIndex, TypeVar
from abc import ABC, abstractmethod
from contextlib import contextmanager
from tabulous._utils import get_config
from tabulous._range import MultiRectRange, RectRange

from qtpy import QtCore, QtWidgets as QtW, QtGui

if TYPE_CHECKING:
    from ._base import QBaseTable
    from ._base._item_model import AbstractDataFrameModel

_Q = TypeVar("_Q", bound=QtCore.QObject)


class _Animation(ABC, Generic[_Q]):
    DURATION = 60

    def __init__(self, parent: _Q):
        self._parent = parent
        self._anim = QtCore.QVariantAnimation(parent)
        self._anim.setDuration(self.DURATION)
        self._anim.valueChanged.connect(self._on_animate)
        self._is_running = False
        self._use_anim = True

    @property
    def should_animate(self) -> bool:
        return self._use_anim and not self._is_running

    @property
    def is_running(self) -> bool:
        return self._is_running

    @property
    def value(self) -> float:
        """Current value of the animation in range [0, 1]."""
        return self._anim.currentValue()

    @abstractmethod
    def _on_animate(self, ratio: float):
        """Animation callback"""

    def set_animate(self, animate: bool):
        self._use_anim = animate
        return self

    @contextmanager
    def using_animation(self, use_anim: bool = True):
        """Context manager to enable/disable animation."""
        _old = self._use_anim
        self._use_anim = use_anim
        try:
            yield None
        finally:
            self._use_anim = _old


class _CellAnimation(_Animation["AbstractDataFrameModel"]):
    def __init__(self, parent: AbstractDataFrameModel):
        super().__init__(parent)
        self._ranges = MultiRectRange([])
        self._draw_region = QtGui.QRegion()
        self._anim.finished.connect(self._finished)

    def _on_animate(self, ratio: float):
        qtable = self._parent.parent()._qtable_view
        qtable.viewport().update(self._draw_region)

    def _finished(self):
        pass


class CellColorAnimation(_CellAnimation):
    DURATION = 240

    def start(self, r: slice, c: slice):
        if not isinstance(r, slice):
            if not isinstance(r, SupportsIndex):
                return
            r = slice(r, r + 1)
        if not isinstance(c, slice):
            if not isinstance(c, SupportsIndex):
                return
            c = slice(c, c + 1)
        if self._use_anim:
            self._ranges = self._ranges.with_slices(r, c)
            new_rect = self._get_rect(self._ranges[-1])
            self._draw_region = self._draw_region.united(new_rect)
            self._is_running = True
            self._anim.setStartValue(0.0)
            self._anim.setEndValue(1.0)
            self._anim.start()
        return None

    def _get_rect(self, rng: RectRange) -> QtCore.QRect:
        """Convert a RectRange into a QRect."""
        rsl, csl = rng.as_iloc()
        r0: int = rsl.start if rsl.start is None else 0
        r1: int = rsl.stop if rsl.stop is None else self._parent.df.shape[0]
        c0: int = csl.start if csl.start is None else 0
        c1: int = csl.stop if csl.stop is None else self._parent.df.shape[1]
        qtable_view = self._parent.parent()._qtable_view
        rect0 = qtable_view.model().index(r0, c0)
        rect1 = qtable_view.model().index(r1, c1)
        return qtable_view.visualRect(rect0) | qtable_view.visualRect(rect1)

    def _finished(self):
        self._is_running = False
        self._ranges = MultiRectRange([])
        self._draw_region = QtGui.QRegion()

    def contains(self, index: QtCore.QModelIndex) -> bool:
        """True if the index is in the current animation range."""
        nr, nc = self._parent.df.shape
        rng = self._ranges.intersection(RectRange(slice(0, nr), slice(0, nc)))
        return (index.row(), index.column()) in rng

    def get_brush(self, size: QtCore.QSize, time: float, bg) -> QtGui.QBrush:
        """Return the brush for the given size and time, with cache support."""
        qtable = self._parent.parent()
        is_qvariant = not isinstance(bg, QtGui.QColor)
        if is_qvariant:
            bg = qtable.palette().color(QtGui.QPalette.ColorRole.Background)
        col = qtable._qtable_view.selectionColor
        length = max(size.width(), size.height())
        dh = 4 / length
        col.setAlpha(96)
        grad = QtGui.QLinearGradient(0, 0, length, length)
        grad.setColorAt(0.0, bg)
        grad.setColorAt(max(time - dh, 0.0), bg)
        grad.setColorAt(time, col)
        grad.setColorAt(min(time + dh, 1.0), bg)
        grad.setColorAt(1.0, bg)
        brush = QtGui.QBrush(grad)
        return brush


class _RowColumnAnimation(_Animation["QBaseTable"]):
    def __init__(self, parent: QBaseTable):
        super().__init__(parent)
        self._anim.setEasingCurve(QtCore.QEasingCurve.Type.InCubic)
        self._index = 0
        self._count = 1

    @abstractmethod
    def _get_header(self) -> QtW.QHeaderView:
        """Get the header object."""

    @abstractmethod
    def _get_span(self, count: int) -> list[int]:
        """Get the span of the section."""

    def run_insert(self, idx: int, count: int, span: int | list[int] | None = None):
        if not self.should_animate:
            return None
        self._is_running = True
        self._index = idx
        self._count = count
        if span is None:
            span = self._get_span(self._count)
        elif not hasattr(span, "__iter__"):
            span = [span] * self._count

        self._init_spans = span
        self._anim.setStartValue(0.0)
        self._anim.setEndValue(1.0)
        self._anim.start()

    def run_remove(self, idx: int, count: int):
        if not self.should_animate:
            return None
        self._is_running = True
        self._index = idx
        self._count = count
        self._init_spans = [
            self._get_header().sectionSize(i)
            for i in range(self._index, self._index + self._count)
        ]
        self._anim.setStartValue(1.0)
        self._anim.setEndValue(0.0)
        self._anim.start()

    def _on_animate(self, ratio: float):
        header = self._get_header()
        for i, span in enumerate(self._init_spans):
            header.resizeSection(self._index + i, int(span * ratio))

    def connect(self, fn: Callable[[], None]):
        """Connect to the finished event of the animation."""
        if not self.should_animate:
            return fn()

        @self._anim.finished.connect
        def _f():
            try:
                with self._parent._mgr.blocked():
                    fn()
            finally:
                self._anim.finished.disconnect(_f)
                self._is_running = False

        return _f


class ColumnAnimation(_RowColumnAnimation):
    def _get_header(self):
        return self._parent._qtable_view.horizontalHeader()

    def _get_span(self, count: int) -> list[int]:
        return [get_config().table.column_size] * count


class RowAnimation(_RowColumnAnimation):
    def _get_header(self):
        return self._parent._qtable_view.verticalHeader()

    def _get_span(self, count: int) -> list[int]:
        return [get_config().table.row_size] * count



from __future__ import annotations
from typing import Any

from qtpy import QtWidgets as QtW
from magicgui.widgets import Widget
from magicgui.backends._qtpy.widgets import QBaseWidget

from tabulous.widgets._table import Table, SpreadSheet


class MagicTable(Widget, Table):
    def __init__(
        self,
        data: Any | None = None,
        *,
        name: str = "",
        editable: bool = False,
        label: str = None,
        tooltip: str | None = None,
        visible: bool | None = None,
        enabled: bool = True,
        gui_only: bool = False,
    ):
        Table.__init__(self, data, name=name, editable=editable)
        super().__init__(
            widget_type=QBaseWidget,
            backend_kwargs={"qwidg": QtW.QWidget},
            name=name,
            label=label,
            tooltip=tooltip,
            visible=visible,
            enabled=enabled,
            gui_only=gui_only,
        )
        mgui_native: QtW.QWidget = self._widget._mgui_get_native_widget()
        mgui_native.setLayout(QtW.QVBoxLayout())
        mgui_native.layout().addWidget(self._qwidget)
        mgui_native.setContentsMargins(0, 0, 0, 0)

    @property
    def native(self):
        return Table.native.fget(self)


class MagicSpreadSheet(Widget, SpreadSheet):
    def __init__(
        self,
        data: Any | None = None,
        *,
        name: str = "",
        editable: bool = True,
        label: str = None,
        tooltip: str | None = None,
        visible: bool | None = None,
        enabled: bool = True,
        gui_only: bool = False,
    ):
        SpreadSheet.__init__(self, data, name=name, editable=editable)
        super().__init__(
            widget_type=QBaseWidget,
            backend_kwargs={"qwidg": QtW.QWidget},
            name=name,
            label=label,
            tooltip=tooltip,
            visible=visible,
            enabled=enabled,
            gui_only=gui_only,
        )
        mgui_native: QtW.QWidget = self._widget._mgui_get_native_widget()
        mgui_native.setLayout(QtW.QVBoxLayout())
        mgui_native.layout().addWidget(self._qwidget)
        mgui_native.setContentsMargins(0, 0, 0, 0)

    @property
    def native(self):
        return SpreadSheet.native.fget(self)



from __future__ import annotations
from typing import Any, Callable, Iterable, TYPE_CHECKING, cast
from qtpy import QtWidgets as QtW
from magicgui import register_type

from magicgui.widgets import (
    Widget,
    Container,
    ComboBox,
    Label,
)
from tabulous.widgets import (
    TableViewerBase,
    TableBase,
    Table,
    SpreadSheet,
)
from tabulous.types import (
    TableColumn,
    TableData,
    TableDataTuple,
    TableInfoInstance,
)

if TYPE_CHECKING:
    import pandas as pd
    from magicgui.widgets import FunctionGui

_DEFAULT_NAME = "Result"


def find_table_viewer_ancestor(
    widget: Widget | QtW.QWidget | TableBase,
) -> TableViewerBase | None:
    from tabulous._qt._mainwindow import _QtMainWidgetBase

    if isinstance(widget, (Widget, TableBase)):
        qwidget = widget.native
    elif isinstance(widget, QtW.QWidget):
        qwidget = widget
    else:
        raise TypeError(f"Cannot use {type(widget)} as an input.")
    qwidget: QtW.QWidget
    while True:
        if isinstance(qwidget, _QtMainWidgetBase):
            qwidget = cast(_QtMainWidgetBase, qwidget)
            return qwidget._table_viewer
        qwidget = qwidget.parent()
        if qwidget is None:
            return None


def find_current_table(widget: Widget | QtW.QWidget) -> TableBase | None:
    viewer = find_table_viewer_ancestor(widget)
    if viewer is None:
        return None
    table = viewer.current_table
    if table is None:
        return None
    return table


def get_any_tables(widget: Widget) -> list[tuple[str, Any]]:
    """Get the list of available tables and the names."""
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return [(t.name, t) for t in v.tables]


def get_tables(widget: Widget) -> list[tuple[str, Any]]:
    """Get the list of available tables and the names."""
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return [(t.name, t) for t in v.tables if isinstance(t, Table)]


def get_spreasheets(widget: Widget) -> list[tuple[str, Any]]:
    """Get the list of available table data and the names."""
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return [(t.name, t) for t in v.tables if isinstance(t, SpreadSheet)]


def get_table_data(widget: Widget) -> list[tuple[str, Any]]:
    """Get the list of available table data and the names."""
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return [(table.name, table.data) for table in v.tables]


def open_viewer(gui, result: TableViewerBase, return_type: type):
    return result.show()


def add_table_to_viewer(
    gui: FunctionGui,
    result: Any,
    return_type: type,
) -> None:
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return
    viewer.add_layer(result, update=True)


def add_table_data_to_viewer(gui: FunctionGui, result: Any, return_type: type) -> None:
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return

    from pandas.core.groupby.generic import DataFrameGroupBy

    if isinstance(result, DataFrameGroupBy):
        viewer.add_groupby(result, name=f"{gui.name}-{_DEFAULT_NAME}", update=True)
    else:
        viewer.add_table(result, name=f"{gui.name}-{_DEFAULT_NAME}", update=True)


def add_table_data_tuple_to_viewer(
    gui: FunctionGui,
    result: tuple,
    return_type: type,
):
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return
    n = len(result)
    if n == 1:
        data = (result[0], f"{gui.name}-{_DEFAULT_NAME}", {})
    elif n == 2:
        if isinstance(result[1], dict):
            name = result[1].pop("name", f"{gui.name}-{_DEFAULT_NAME}")
            data = (result[0], name, result[1])
        else:
            data = (result[0], result[1], {})
    elif n == 3:
        data = result
    else:
        raise ValueError(f"Length of TableDataTuple must be < 4, got {n}.")
    viewer.add_table(data[0], name=data[1], **data[2], update=True)


register_type(
    TableViewerBase,
    return_callback=open_viewer,
    bind=find_table_viewer_ancestor,
)
register_type(TableBase, return_callback=add_table_to_viewer, choices=get_any_tables)
register_type(Table, return_callback=add_table_to_viewer, choices=get_tables)
register_type(SpreadSheet, return_callback=add_table_to_viewer, choices=get_spreasheets)
register_type(
    TableData,
    return_callback=add_table_data_to_viewer,
    choices=get_table_data,
    nullable=False,
)
register_type(TableDataTuple, return_callback=add_table_data_tuple_to_viewer)

# Widget
class ColumnChoice(Container):
    def __init__(
        self,
        data_choices: Iterable[pd.DataFrame]
        | Callable[[Widget], Iterable[pd.DataFrame]],
        value=None,
        **kwargs,
    ):
        self._dataframe_choices = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_choices = ComboBox(choices=self._get_available_columns)
        _label_l = Label(value='["')
        _label_l.max_width = 24
        _label_r = Label(value='"]')
        _label_r.max_width = 24
        super().__init__(
            layout="horizontal",
            widgets=[self._dataframe_choices, _label_l, self._column_choices, _label_r],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_choices.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: pd.DataFrame = self._dataframe_choices.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        self._column_choices.choices = cols
        return None

    @property
    def value(self) -> pd.Series:
        df = self._dataframe_choices.value
        return df[self._column_choices.value]


register_type(
    TableColumn,
    widget_type=ColumnChoice,
    return_callback=add_table_data_to_viewer,
    data_choices=get_table_data,
    nullable=False,
)


class ColumnNameChoice(Container):
    """
    A container widget with a DataFrame selection and multiple column name selections.

    This widget is composed of two or more ComboBox widgets. The top one is to choose a
    DataFrame and the rest are to choose column names from the DataFrame. When the
    DataFrame selection changed, the column name selections will also changed
    accordingly.
    """

    def __init__(
        self,
        data_choices: Iterable[pd.DataFrame]
        | Callable[[Widget], Iterable[pd.DataFrame]],
        column_choice_names: Iterable[str],
        value=None,
        **kwargs,
    ):
        self._dataframe_choices = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_names: list[ComboBox] = []
        for cn in column_choice_names:
            self._column_names.append(
                ComboBox(choices=self._get_available_columns, name=cn, nullable=True)
            )
        self._child_container = Container(widgets=self._column_names, layout="vertical")
        self._child_container.margins = (0, 0, 0, 0)
        super().__init__(
            layout="vertical",
            widgets=[self._dataframe_choices, self._child_container],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_choices.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: pd.DataFrame = self._dataframe_choices.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        for cbox in self._column_names:
            cbox.choices = cols
        return None

    @property
    def value(self) -> tuple[pd.DataFrame, list[str]]:
        df = self._dataframe_choices.value
        colnames = [cbox.value for cbox in self._column_names]
        return (df, colnames)


register_type(
    TableInfoInstance, widget_type=ColumnNameChoice, data_choices=get_table_data
)



from __future__ import annotations
from typing import Iterator, TYPE_CHECKING, NamedTuple, Generic, TypeVar, Union
import weakref
import logging
from dataclasses import dataclass

import numpy as np
import pandas as pd

# NOTE: Axes should be imported here!
from tabulous.widgets import TableBase
from tabulous._selection_op import SelectionOperator
from tabulous._magicgui import Axes

if TYPE_CHECKING:
    from matplotlib.artist import Artist
    from matplotlib.collections import PathCollection, PolyCollection
    from matplotlib.lines import Line2D
    from matplotlib.patches import Polygon
    from matplotlib.container import BarContainer
    from tabulous._qt._plot import QtMplPlotCanvas

logger = logging.getLogger(__name__)
_T = TypeVar("_T", bound="Artist")


class PlotRef(NamedTuple):
    """Reference objects used to update plots."""

    widget: weakref.ReferenceType[QtMplPlotCanvas]
    artists: list[weakref.ReferenceType[Artist]]

    def deref(self) -> tuple[QtMplPlotCanvas, list[Artist]]:
        """Dereference the weak references."""
        _widget = self.widget()
        _artists = [a() for a in self.artists]
        if _widget is None or None in _artists:
            raise RuntimeError("Widget and/or artists have been deleted")
        return _widget, _artists


class AbstractDataModel(Generic[_T]):
    ax: Axes
    table: TableBase
    ref = False

    def update_ax(self, *args, **kwargs) -> _T:
        raise NotImplementedError()

    def update_artist(self, artist: _T, *args: pd.Series):
        raise NotImplementedError()

    def add_data(self):
        raise NotImplementedError()

    def update_data(self, artists: list[_T], mpl_widget: QtMplPlotCanvas) -> bool:
        """
        Update the data of the artist.
        Return True if the data is successfully updated.
        """
        try:
            for i, data in enumerate(self._iter_data()):
                self.update_artist(artists[i], *data)
            mpl_widget.draw()
        except RuntimeError as e:
            if str(e).startswith("wrapped C/C++ object of"):
                # Qt widget is deleted.
                return False

        return True

    def _get_reactive_ranges(self) -> list[tuple[slice, slice]]:
        raise NotImplementedError()

    def _iter_data(self) -> Iterator[tuple[pd.Series, ...]]:
        """Iterate over the data to be plotted."""
        raise NotImplementedError()


class YDataModel(AbstractDataModel[_T]):
    ax: Axes
    y_selection: SelectionOperator
    table: TableBase

    label_selection = None  # default
    ref = False

    def add_data(self):
        _mpl_widget = weakref.ref(self.table.plt.gcw())
        _artist_refs: list[weakref.ReferenceType[_T]] = []
        for (y,) in self._iter_data():
            label_name = y.name
            artist = self.update_ax(y, label=label_name)
            # _artist_refs.append(weakref.ref(artist))  TODO: cannot weakref BarContainer

        if not self.ref:
            # if plot does not refer the table data, there's nothing to be done
            return

        # NOTE: Unreachable below
        plot_ref = PlotRef(_mpl_widget, _artist_refs)

        def _on_data_updated():
            # when any of the data is updated, reset the scatter offsets
            try:
                _plt, _artists = plot_ref.deref()
            except RuntimeError:
                updated = False
            else:
                updated = self.update_data(_artists, _plt)
            if not updated:
                self.table.events.data.disconnect(_on_data_updated)
                logger.debug("Disconnecting scatter plot.")

        reactive_ranges = self._get_reactive_ranges()
        self.table.events.data.mloc(reactive_ranges).connect(_on_data_updated)
        return None

    def _get_reactive_ranges(self) -> list[tuple[slice, slice]]:
        data = self.table.data
        yslice = self.y_selection.as_iloc_slices(data)
        reactive_ranges = [yslice]
        return reactive_ranges

    def _iter_data(self) -> Iterator[tuple[pd.Series]]:
        """Iterate over the data to be plotted."""
        data = self.table.data
        if self.y_selection is None:
            raise ValueError("Y must be set.")

        yslice = self.y_selection.as_iloc_slices(data)
        ydata_all = data.iloc[yslice]

        if self.label_selection is None:
            for _, ydata in ydata_all.items():
                yield (ydata,)
        else:
            ldata = get_column(self.label_selection, data)
            lable_unique = ldata.unique()

            for l in lable_unique:
                spec = ldata == l
                for _, ydata in ydata_all[spec].items():
                    yield (ydata,)


class XYDataModel(AbstractDataModel[_T]):
    ax: Axes
    x_selection: SelectionOperator | None
    y_selection: SelectionOperator
    table: TableBase

    label_selection = None  # default
    ref = False

    def add_data(self):
        _mpl_widget = weakref.ref(self.table.plt.gcw())
        _artists: list[_T] = []
        for x, y in self._iter_data():
            label_name = y.name
            artist = self.update_ax(x, y, label=label_name)
            _artists.append(artist)

        if not self.ref:
            # if plot does not refer the table data, there's nothing to be done
            return

        _artist_refs: list[weakref.ReferenceType[_T]] = []
        for artist in _artists:
            _artist_refs.append(weakref.ref(artist))

        plot_ref = PlotRef(_mpl_widget, _artist_refs)

        def _on_data_updated():
            # when any of the data is updated, reset the scatter offsets
            try:
                _plt, _artists = plot_ref.deref()
            except RuntimeError:
                updated = False
            else:
                updated = self.update_data(_artists, _plt)
            if not updated:
                self.table.events.data.disconnect(_on_data_updated)
                logger.debug("Disconnecting scatter plot.")

        reactive_ranges = self._get_reactive_ranges()
        self.table.events.data.mloc(reactive_ranges).connect(_on_data_updated)
        return None

    def _get_reactive_ranges(self) -> list[tuple[slice, slice]]:
        data = self.table.data
        yslice = self.y_selection.as_iloc_slices(data)
        reactive_ranges = [yslice]

        if self.x_selection is not None:
            xslice = self.x_selection.as_iloc_slices(data)
            reactive_ranges.append(xslice)

        return reactive_ranges

    def _iter_data(self) -> Iterator[tuple[pd.Series, pd.Series]]:
        """Iterate over the data to be plotted."""
        data = self.table.data
        if self.y_selection is None:
            raise ValueError("Y must be set.")

        yslice = self.y_selection.as_iloc_slices(data)
        ydata_all = data.iloc[yslice]
        # TODO: support row vector
        if self.x_selection is None:
            xdata = pd.Series(np.arange(len(ydata_all)), name="X")
        else:
            xdata = get_column(self.x_selection, data)

        if self.label_selection is None:
            for _, ydata in ydata_all.items():
                yield xdata, ydata
        else:
            ldata = get_column(self.label_selection, data)
            lable_unique = ldata.unique()

            for l in lable_unique:
                spec = ldata == l
                xdata_subset = xdata[spec]
                for _, ydata in ydata_all[spec].items():
                    yield xdata_subset, ydata


class XYYDataModel(AbstractDataModel[_T]):
    ax: Axes
    x_selection: SelectionOperator | None
    y0_selection: SelectionOperator
    y1_selection: SelectionOperator
    table: TableBase

    label_selection = None  # default
    ref = False

    def add_data(self):
        _mpl_widget = weakref.ref(self.table.plt.gcw())
        _artists: list[_T] = []
        for x, y0, y1 in self._iter_data():
            label_name = y0.name
            artist = self.update_ax(x, y0, y1, label=label_name)
            _artists.append(artist)

        if not self.ref:
            # if plot does not refer the table data, there's nothing to be done
            return

        _artist_refs: list[weakref.ReferenceType[_T]] = []
        for artist in _artists:
            _artist_refs.append(weakref.ref(artist))

        plot_ref = PlotRef(_mpl_widget, _artist_refs)

        def _on_data_updated():
            # when any of the data is updated, reset the scatter offsets
            try:
                _plt, _artists = plot_ref.deref()
            except RuntimeError:
                updated = False
            else:
                updated = self.update_data(_artists, _plt)
            if not updated:
                self.table.events.data.disconnect(_on_data_updated)
                logger.debug("Disconnecting scatter plot.")

        reactive_ranges = self._get_reactive_ranges()
        self.table.events.data.mloc(reactive_ranges).connect(_on_data_updated)
        return None

    def _get_reactive_ranges(self) -> list[tuple[slice, slice]]:
        data = self.table.data
        y0slice = self.y0_selection.as_iloc_slices(data)
        y1slice = self.y1_selection.as_iloc_slices(data)
        reactive_ranges = [y0slice, y1slice]

        if self.x_selection is not None:
            xslice = self.x_selection.as_iloc_slices(data)
            reactive_ranges.append(xslice)

        return reactive_ranges

    def _iter_data(self) -> Iterator[tuple[pd.Series, pd.Series, pd.Series]]:
        """Iterate over the data to be plotted."""
        data = self.table.data
        if self.y0_selection is None or self.y1_selection is None:
            raise ValueError("Y0 and Y1 must be set.")

        y0data = get_column(self.y0_selection, data)
        y1data = get_column(self.y1_selection, data)
        if len(y0data) != len(y1data):
            raise ValueError("Y0 and Y1 must have the same length.")
        if self.x_selection is None:
            xdata = pd.Series(np.arange(len(y0data)), name="X")
        else:
            xdata = get_column(self.x_selection, data)

        if self.label_selection is None:
            yield xdata, y0data, y1data
        else:
            ldata = get_column(self.label_selection, data)
            lable_unique = ldata.unique()

            for l in lable_unique:
                spec = ldata == l
                xdata_subset = xdata[spec]
                yield xdata_subset, y0data[spec], y1data[spec]


@dataclass
class PlotModel(XYDataModel["Line2D"]):
    ax: Axes
    x_selection: SelectionOperator | None
    y_selection: SelectionOperator
    table: TableBase
    alpha: float = 1.0
    ref: bool = False

    def update_ax(self, x, y, label=None) -> Line2D:
        return self.ax.plot(x, y, alpha=self.alpha, label=label, picker=True)[0]

    def update_artist(self, artist: Line2D, x: pd.Series, y: pd.Series):
        return artist.set_data(x, y)


@dataclass
class BarModel(XYDataModel["BarContainer"]):
    ax: Axes
    x_selection: SelectionOperator | None
    y_selection: SelectionOperator
    table: TableBase
    alpha: float = 1.0
    ref: bool = False

    def update_ax(self, x, y, label=None) -> BarContainer:
        return self.ax.bar(x, y, alpha=self.alpha, label=label, picker=True)

    def update_artist(self, artist: BarContainer, x: pd.Series, y: pd.Series):
        for patch in artist.patches:
            width = patch.get_width()
            patch.set_x(x - width / 2)
            patch.set_height(y)
        return None


@dataclass
class FillBetweenModel(XYYDataModel["PolyCollection"]):
    ax: Axes
    x_selection: SelectionOperator | None
    y0_selection: SelectionOperator
    y1_selection: SelectionOperator
    table: TableBase
    alpha: float = 1.0
    ref: bool = False

    def update_ax(self, x, y0, y1, label=None) -> PolyCollection:
        return self.ax.fill_between(
            x, y0, y1, alpha=self.alpha, label=label, picker=True
        )

    def update_artist(
        self, artist: PolyCollection, x: pd.Series, y0: pd.Series, y1: pd.Series
    ):
        new_verts = np.concatenate(
            [np.stack([x, y0], axis=1), np.stack([x, y1], axis=1)[::-1]],
            axis=0,
        )
        artist.set_verts(new_verts[np.newaxis])
        return None


@dataclass
class FillBetweenXModel(FillBetweenModel):
    def update_ax(self, x, y0, y1, label=None) -> PolyCollection:
        return self.ax.fill_betweenx(
            x, y0, y1, alpha=self.alpha, label=label, picker=True
        )

    def update_artist(
        self, artist: PolyCollection, x: pd.Series, y0: pd.Series, y1: pd.Series
    ):
        new_verts = np.concatenate(
            [np.stack([y0, x], axis=1), np.stack([y1, x], axis=1)[::-1]],
            axis=0,
        )
        artist.set_verts(new_verts[np.newaxis])
        return None


@dataclass
class ScatterModel(XYDataModel["PathCollection"]):
    ax: Axes
    x_selection: SelectionOperator | None
    y_selection: SelectionOperator
    table: TableBase
    label_selection: SelectionOperator | None = None
    alpha: float = 1.0
    ref: bool = False

    def update_ax(self, x, y, label=None) -> PathCollection:
        return self.ax.scatter(x, y, alpha=self.alpha, label=label, picker=True)

    def update_artist(self, artist: PathCollection, x: pd.Series, y: pd.Series):
        return artist.set_offsets(np.stack([x, y], axis=1))


@dataclass
class HistModel(YDataModel[Union["BarContainer", "Polygon"]]):
    ax: Axes
    y_selection: SelectionOperator
    bins: int
    table: TableBase
    range: tuple[float, float] | None = None
    label_selection: SelectionOperator | None = None
    alpha: float = 1.0
    density: bool = False
    histtype: str = "bar"
    ref: bool = False

    def update_ax(self, y, label=None) -> Union[BarContainer, Polygon]:
        artist = self.ax.hist(
            y,
            alpha=self.alpha,
            label=label,
            bins=self.bins,
            density=self.density,
            range=self.range,
            histtype=self.histtype,
            picker=True,
        )[2]
        if self.histtype in ("step", "stepfilled"):
            return artist[0]
        else:
            return artist

    def update_artist(self, artist: Union[BarContainer, Polygon], y: pd.Series):
        raise NotImplementedError()


def get_column(selection: SelectionOperator, df: pd.DataFrame) -> pd.Series:
    sl = selection.as_iloc_slices(df)
    data = df.iloc[sl]
    if data.shape[1] != 1:
        raise ValueError("Label must be a single column.")
    return data.iloc[:, 0]



from __future__ import annotations
from typing import Any, Callable, Hashable, TYPE_CHECKING, Iterable, cast
import warnings
from qtpy import QtCore, QtGui, QtWidgets as QtW
from qtpy.QtCore import Qt, Signal
import pandas as pd

from tabulous._dtype import isna
from tabulous.color import normalize_color, ColorType
from tabulous._text_formatter import DefaultFormatter
from tabulous._map_model import TableMapping
from tabulous._utils import get_config
from tabulous._qt._table._animation import CellColorAnimation

if TYPE_CHECKING:
    from ._table_base import QBaseTable

# https://ymt-lab.com/post/2020/pyqt5-qtableview-pandas-qabstractitemmodel/

_READ_ONLY = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable


class AbstractDataFrameModel(QtCore.QAbstractTableModel):
    """Table model for data frame."""

    dataEdited = Signal(int, int, object)
    _FORMAT_ERROR = "<Format Error>"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._df = pd.DataFrame([])

        self._editable = False
        self._foreground_colormap: dict[Hashable, Callable[[Any], ColorType]] = {}
        self._background_colormap: dict[Hashable, Callable[[Any], ColorType]] = {}
        self._text_formatter: dict[Hashable, Callable[[Any], str]] = {}
        self._validator: dict[Hashable, Callable[[Any], None]] = {}

        self._data_role_map = {
            Qt.ItemDataRole.DisplayRole: self._data_display,
            Qt.ItemDataRole.EditRole: self._data_edit,
            Qt.ItemDataRole.TextColorRole: self._data_text_color,
            Qt.ItemDataRole.ToolTipRole: self._data_tooltip,
            Qt.ItemDataRole.BackgroundColorRole: self._data_background_color_rendered,
            Qt.ItemDataRole.DecorationRole: self._data_decoration,
            Qt.ItemDataRole.SizeHintRole: self._data_size_hint,
        }

        self._decorations: TableMapping[tuple[QtGui.QPixmap, str]] = TableMapping()
        self._background_color_anim = CellColorAnimation(self)

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    def updateValue(self, r, c, val):
        # pandas warns but no problem
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._df.iloc[r, c] = val

    def data(
        self,
        index: QtCore.QModelIndex,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if not index.isValid():
            return QtCore.QVariant()
        if map := self._data_role_map.get(role, None):
            return map(index)
        return QtCore.QVariant()

    def _data_display(self, index: QtCore.QModelIndex):
        """Display role."""
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            val = df.iat[r, c]
            colname = df.columns[c]
            if isna(val):
                text = "NA"
            elif mapper := self._text_formatter.get(colname, None):
                try:
                    text = str(mapper(self.parent().convertValue(c, val)))
                except Exception:
                    text = self._FORMAT_ERROR
            else:
                fmt = DefaultFormatter(df.dtypes[colname])
                text = fmt(val)
            return text
        return QtCore.QVariant()

    def _data_edit(self, index: QtCore.QModelIndex):
        """Edit role."""
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            base_table = self.parent()
            if ref_expr := base_table._get_ref_expr(r, c):
                return get_config().cell.ref_prefix + ref_expr

            val = df.iat[r, c]
            if isna(val):
                text = "NA"
            else:
                text = str(val)
            return text
        return QtCore.QVariant()

    def _data_text_color(self, index: QtCore.QModelIndex):
        if not self._foreground_colormap:
            return QtCore.QVariant()
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            colname = df.columns[c]
            val = df.iat[r, c]
            if mapper := self._foreground_colormap.get(colname, None):
                # If mapper is given for the column, call it.
                try:
                    col = mapper(val)
                    if col is None:
                        if isna(val):
                            return QtGui.QColor(Qt.GlobalColor.gray)
                        return QtCore.QVariant()
                    rgba = normalize_color(col)
                except Exception as e:
                    # since this method is called many times, errorous function should
                    # be deleted from the mapper.
                    self._foreground_colormap.pop(colname)
                    raise e
                return QtGui.QColor(*rgba)
            if isna(val):
                return QtGui.QColor(Qt.GlobalColor.gray)
        return QtCore.QVariant()

    def _data_tooltip(self, index: QtCore.QModelIndex):
        r, c = index.row(), index.column()
        if r < self.df.shape[0] and c < self.df.shape[1]:
            val = self.df.iat[r, c]
            dtype = self.df.dtypes.values[c]
            if ref_expr := self.parent()._get_ref_expr_by_dest(r, c):
                ref = f"\nExpr: {ref_expr}"
            else:
                ref = ""
            if dtype != object:
                return f"{val!r} (dtype: {dtype}){ref}"
            else:
                return f"{val!r} (dtype: {dtype}; type: {type(val).__name__}){ref}"
        return QtCore.QVariant()

    def _data_background_color(self, index: QtCore.QModelIndex):
        if not self._background_colormap:
            return QtCore.QVariant()
        r, c = index.row(), index.column()
        df = self.df
        if r < df.shape[0] and c < df.shape[1]:
            colname = df.columns[c]
            if mapper := self._background_colormap.get(colname, None):
                val = df.iat[r, c]
                try:
                    col = mapper(val)
                    if col is None:
                        return QtCore.QVariant()
                    rgba = normalize_color(col)
                except Exception as e:
                    # since this method is called many times, errorous function should
                    # be deleted from the mapper.
                    self._background_colormap.pop(colname)
                    raise e
                return QtGui.QColor(*rgba)
        return QtCore.QVariant()

    def _data_background_color_rendered(self, index: QtCore.QModelIndex):
        col = self._data_background_color(index)
        anim = self._background_color_anim
        if not anim.contains(index):
            return col
        return anim.get_brush(self._data_size_hint(index), anim.value, col)

    def _data_decoration(self, index: QtCore.QModelIndex):
        r, c = index.row(), index.column()
        if content := self._decorations.get((r, c), None):
            return content[0]
        return QtCore.QVariant()

    def _data_size_hint(self, index: QtCore.QModelIndex):
        r, c = index.row(), index.column()
        tv = self.parent()._qtable_view
        vsize = tv.verticalHeader()._section_sizes
        hsize = tv.horizontalHeader()._section_sizes
        if r < len(vsize):
            lr = tv.verticalHeader()._section_sizes[r]
        else:
            lr = get_config().table.row_size
        if c < len(hsize):
            lc = tv.horizontalHeader()._section_sizes[c]
        else:
            lc = get_config().table.column_size
        zoom = self.parent().zoom()
        return QtCore.QSize(int(lc * zoom), int(lr * zoom))

    def set_cell_label(self, index: QtCore.QModelIndex, text: str | None):
        if text is None or text == "":
            self._decorations.pop((index.row(), index.column()), None)
        else:
            qlabel = QtW.QLabel(text, self.parent())
            qlabel.setStyleSheet("background-color: transparent; color: gray;")
            qlabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
            font = self.parent()._qtable_view.font()
            font.setBold(True)
            qlabel.setFont(font)
            rect = QtGui.QFontMetrics(font).boundingRect(text)
            qlabel.resize(rect.size())
            pixmap = qlabel.grab()
            qlabel.deleteLater()
            self._decorations[(index.row(), index.column())] = pixmap, text
        qtable_view = self.parent()._qtable_view
        qtable_view.update(index)
        return None

    def flags(self, index):
        if self._editable:
            return Qt.ItemFlag.ItemIsEditable | _READ_ONLY
        else:
            return _READ_ONLY

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.ItemDataRole.DisplayRole:
                if section >= self.df.columns.size:
                    return None
                if self.df.columns.nlevels == 1:
                    text = str(self.df.columns[section])
                else:
                    text = "\n".join(map(str, self.df.columns[section]))
                return text
            elif role == Qt.ItemDataRole.ToolTipRole:
                if section < self.df.columns.size:
                    return self._column_tooltip(section)
                return None

        if orientation == Qt.Orientation.Vertical:
            if role == Qt.ItemDataRole.DisplayRole:
                if section >= self.df.index.size:
                    return None
                text = str(self.df.index[section]) + " "
                return text
            elif role == Qt.ItemDataRole.ToolTipRole:
                if section < self.df.index.size:
                    return str(self.df.index[section])
                return None

    def _column_tooltip(self, section: int):
        name = self.df.columns[section]
        dtype = self.df.dtypes.values[section]
        return f"{name} (dtype: {dtype})"

    def rename_column(self, old_name: str, new_name: str):
        """Fix keys to match new column names."""
        if background_colormap := self._background_colormap.pop(old_name, None):
            self._background_colormap[new_name] = background_colormap
        if foreground_colormap := self._foreground_colormap.pop(old_name, None):
            self._foreground_colormap[new_name] = foreground_colormap
        if text_formatter := self._text_formatter.pop(old_name, None):
            self._text_formatter[new_name] = text_formatter
        if validator := self._validator.pop(old_name, None):
            self._validator[new_name] = validator
        return None

    def setData(self, index: QtCore.QModelIndex, value, role) -> bool:
        if not index.isValid():
            return False
        if role != Qt.ItemDataRole.EditRole:
            return False
        r, c = index.row(), index.column()
        self.dataEdited.emit(r, c, value)
        return True

    def setShape(self, nrow: int, ncol: int):
        """Set table shape."""
        r0, c0 = self.rowCount(), self.columnCount()
        dr = nrow - r0
        dc = ncol - c0

        # Adjust rows
        parent = QtCore.QModelIndex()
        if dr > 0:
            self.beginInsertRows(parent, r0, r0 + dr - 1)
            self.insertRows(r0, dr, parent)
            self.endInsertRows()
        elif dr < 0:
            self.beginRemoveRows(parent, r0 + dr, r0 - 1)
            self.removeRows(r0 + dr, -dr, parent)
            self.endRemoveRows()

        # Adjust columns
        if dc > 0:
            self.beginInsertColumns(parent, c0, c0 + dc - 1)
            self.insertColumns(c0, dc, parent)
            self.endInsertColumns()
        elif dc < 0:
            self.beginRemoveColumns(parent, c0 + dc, c0 - 1)
            self.removeColumns(c0 + dc, -dc, parent)
            self.endRemoveColumns()

        return None

    def insertColumns(
        self,
        column: int,
        count: int,
        parent: QtCore.QModelIndex = None,
        span: int | Iterable[int] | None = None,
    ) -> bool:
        self._decorations.insert_columns(column, count)
        if span is None:
            span = self.data(self.index(0, column - 1), Qt.ItemDataRole.SizeHintRole)
            if isinstance(span, QtCore.QSize):
                span = cast(QtCore.QSize, span).width()
            else:
                span = get_config().table.column_size
        header = self.parent()._qtable_view.horizontalHeader()
        if hasattr(span, "__iter__"):
            for s, r in zip(span, range(column, column + count)):
                header.insertSection(r, 1, s)
        else:
            header.insertSection(column, count, span)
        return super().insertColumns(column, count, parent)

    def removeColumns(
        self, column: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        self._decorations.remove_columns(column, count)
        self.parent()._qtable_view.horizontalHeader().removeSection(column, count)
        return super().removeColumns(column, count, parent)

    def insertRows(
        self,
        row: int,
        count: int,
        parent: QtCore.QModelIndex = None,
        span: int | Iterable[int] | None = None,
    ) -> bool:
        self._decorations.insert_rows(row, count)
        if span is None:
            span = self.data(self.index(0, row - 1), Qt.ItemDataRole.SizeHintRole)
            if isinstance(span, QtCore.QSize):
                span = cast(QtCore.QSize, span).height()
            else:
                span = get_config().table.row_size
        header = self.parent()._qtable_view.verticalHeader()
        if hasattr(span, "__iter__"):
            for s, r in zip(span, range(row, row + count)):
                header.insertSection(r, 1, s)
        else:
            header.insertSection(row, count, span)
        return super().insertRows(row, count, parent)

    def removeRows(
        self, row: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        self._decorations.remove_rows(row, count)
        self.parent()._qtable_view.verticalHeader().removeSection(row, count)
        return super().removeRows(row, count, parent)

    if TYPE_CHECKING:

        def parent(self) -> QBaseTable:
            ...


class DataFrameModel(AbstractDataFrameModel):
    """A concrete model for a pandas DataFrame."""

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, data: pd.DataFrame):
        if data is self._df:
            return
        self.setShape(*data.shape)
        self._df = data

    def rowCount(self, parent=None):
        return self.df.shape[0]

    def columnCount(self, parent=None):
        return self.df.shape[1]



from __future__ import annotations

from pathlib import Path
from qtpy.QtWidgets import QApplication
from qtpy.QtCore import Qt
from qtpy import PYQT5, QtGui

ICON_PATH = Path(__file__).parent / "_icons"

APPLICATION = None


def gui_qt():
    """Call "%gui qt" magic."""
    try:
        from IPython import get_ipython
    except ImportError:

        def get_ipython():
            return False

    shell = get_ipython()

    if shell and shell.active_eventloop != "qt":
        shell.enable_gui("qt")
    return None


def gui_qt_is_active() -> bool:
    """True only if "%gui qt" magic is called in ipython kernel."""
    try:
        from IPython import get_ipython
    except ImportError:

        def get_ipython():
            return False

    shell = get_ipython()

    return shell and shell.active_eventloop == "qt"


def get_app():
    """Get QApplication."""
    global APPLICATION
    gui_qt()
    app = QApplication.instance()
    if app is None:
        if PYQT5:
            QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
            QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)
        app = QApplication([])

    if app.windowIcon().isNull():
        app.setWindowIcon(QtGui.QIcon(str(ICON_PATH / "window_icon.png")))
    app.setApplicationName("tabulous")
    APPLICATION = app
    return app


def run_app():
    """Start the event loop."""
    if not gui_qt_is_active():
        from tabulous.exceptions import ExceptionHandler

        with ExceptionHandler(hook=_excepthook) as _:
            get_app().exec_()
        return None


def _excepthook(exc_type: type[Exception], exc_value: Exception, exc_traceback):
    """Exception hook used during application execution."""
    from ._traceback import QtErrorMessageBox
    from ._mainwindow import QMainWindow

    viewer = QMainWindow.currentViewer()
    QtErrorMessageBox.raise_(exc_value, parent=viewer.native)
    return None



from __future__ import annotations
from typing import Any

from magicgui import register_type
from magicgui.widgets import (
    Container,
    PushButton,
    LineEdit,
)

from tabulous.widgets import TableBase, SpreadSheet
from tabulous._selection_op import (
    SelectionOperator,
    ILocSelOp,
    parse,
    construct,
)

from ._register import find_current_table


class SelectionWidget(Container):
    """A container widget for a table selection."""

    def __init__(
        self,
        value: Any = None,
        nullable=False,
        format: str = "iloc",
        allow_out_of_bounds: bool = False,
        **kwargs,
    ):
        self._line = LineEdit(tooltip="Selection string (e.g. 'df.iloc[2:4, 3:]')")
        self._btn = PushButton(
            text="Read selection", tooltip="Read current table selection."
        )
        super().__init__(layout="horizontal", widgets=[self._line, self._btn], **kwargs)
        self.margins = (0, 0, 0, 0)
        self._line.changed.disconnect()
        self._btn.changed.disconnect()
        self._line.changed.connect(self.changed.emit(self._line.value))
        self._btn.changed.connect(lambda: self._read_selection())

        self._format = format
        self._allow_out_of_bounds = allow_out_of_bounds

        if isinstance(value, (str, SelectionOperator, tuple)):
            self.value = value

    @property
    def value(self) -> SelectionOperator | None:
        """Get selection operator that represents current selection."""
        text = self._line.value
        if text:
            return parse(text, df_expr="df")
        return None

    @value.setter
    def value(self, val: str | SelectionOperator) -> None:
        if isinstance(val, str):
            if val:
                text = parse(val, df_expr="df").fmt()
            else:
                text = ""
        elif isinstance(val, SelectionOperator):
            text = val.fmt()
        elif isinstance(val, tuple):
            text = ILocSelOp(*val).fmt()
        elif val is None:
            text = ""
        else:
            raise TypeError(f"Invalid type for value: {type(val)}")
        self._line.value = text
        return None

    @property
    def format(self) -> str:
        return self._format

    def as_iloc(self) -> tuple[slice, slice]:
        """Return current value as a indexer for ``iloc`` method."""
        df = self._find_table().data_shown
        return self.value.as_iloc(df)

    def as_iloc_slices(self) -> tuple[slice, slice]:
        """Return current value as slices for ``iloc`` method."""
        df = self._find_table().data_shown
        return self.value.as_iloc_slices(df)

    def _find_table(self) -> TableBase:
        table = find_current_table(self)
        if table is None:
            raise ValueError("No table found.")
        return table

    def _read_selection(self, table: TableBase | None = None):
        if table is None:
            table = self._find_table()

        sels = table.selections
        if len(sels) > 1:
            raise ValueError("More than one selection is given.")
        sel = sels[0]

        qwidget = table.native
        column_selected = qwidget._qtable_view._selection_model._col_selection_indices
        if isinstance(table, SpreadSheet):
            df = table.native._data_raw
        else:
            df = table.data_shown
        _selop = construct(
            *sel,
            df,
            method=self.format,
            column_selected=column_selected,
            allow_out_of_bounds=self._allow_out_of_bounds,
        )
        self._line.value = _selop.fmt()
        self.changed.emit(_selop)
        return None


register_type(SelectionOperator, widget_type=SelectionWidget)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Any
import re
from ._base import Component

if TYPE_CHECKING:
    from tabulous.widgets._keymap_abc import SupportKeyMap

# e.g. f() takes 0 positional arguments but 1 was given
_PATTERN = re.compile(r".*takes 0 positional arguments but (\d+) w.+ given")


class KeyMap(Component["SupportKeyMap"]):
    def register(
        self,
        key: str,
        func: Callable[[SupportKeyMap], Any] | None = None,
        overwrite: bool = False,
    ):
        def wrapper(f):
            def _inner(*_):
                try:
                    out = f(self.parent)
                except TypeError as e:
                    if _PATTERN.match(str(e)):
                        out = f()
                    else:
                        raise e
                return out

            return self.parent._qwidget._keymap.bind(key, _inner, overwrite=overwrite)

        return wrapper if func is None else wrapper(func)

    def unregister(self, key: str) -> None:
        self.parent._qwidget._keymap.unbind(key)

    def press_key(self, key: str) -> None:
        self.parent._qwidget._keymap.press_key(key)
        return None


def bind(self: KeyMap, *args, **kwargs):
    import warnings

    warnings.warn(
        "Keycombo registration using `keymap.bind` is deprecated. Use "
        "`keymap.register` instead.",
        DeprecationWarning,
    )
    return self.parent._qwidget._keymap.bind(*args, **kwargs)


KeyMap.bind = bind  # backward compatibility



from __future__ import annotations

from typing import Union, TYPE_CHECKING
from pathlib import Path

if TYPE_CHECKING:
    import pandas as pd

PathLike = Union[str, Path, bytes]


def open_file(path: PathLike) -> pd.DataFrame | dict[str, pd.DataFrame]:
    """
    Read a table data and add to the viewer.

    Parameters
    ----------
    path : path like
        File path.
    """
    import pandas as pd

    path = Path(path)
    suf = path.suffix

    if suf in (".csv", ".txt", ".dat"):
        df = pd.read_csv(path, index_col=_get_index_col(path))
    elif suf in (".xlsx", ".xls", ".xlsb", ".xlsm", ".xltm", "xltx", ".xml"):
        df: dict[str, pd.DataFrame] = pd.read_excel(path, sheet_name=None)
    elif suf in (".parquet", ".pq"):
        df = pd.read_parquet(path)
    else:
        raise ValueError(f"Extension {suf!r} not supported.")
    return df


def save_file(path: PathLike, df: pd.DataFrame) -> None:
    """Save current table."""
    path = Path(path)
    suf = path.suffix
    if suf in (".csv", ".txt", ".dat"):
        df.to_csv(path)
    elif suf in (".xlsx", ".xls", "xml"):
        df.to_excel(path)
    elif suf in (".html",):
        df.to_html(path)
    elif suf in (".parquet", ".pq"):
        df.to_parquet(path)
    else:
        raise ValueError(f"Extension {suf} not supported.")


def _get_index_col(path: PathLike, sep=",") -> int | None:
    index_col = None
    try:
        with open(path) as f:
            first_char = f.read(1)
        if first_char == sep:
            index_col = 0
    except Exception:
        pass
    return index_col



from __future__ import annotations

from types import FunctionType
from typing import (
    Any,
    Callable,
    TYPE_CHECKING,
)
from functools import wraps, partial

if TYPE_CHECKING:
    from typing_extensions import Self, ParamSpec
    from ._keymap_objects import QtKeys

    _P = ParamSpec("_P")


class BoundCallback(partial):
    desc: str

    def __new__(
        cls,
        func: Callable[_P, Any],
        /,
        *,
        desc: str = "",
        keys: QtKeys | None = None,
        kwargs: dict[str, Any] = {},
    ) -> Self:
        kwargs = kwargs.copy()

        import inspect

        try:
            sig = inspect.signature(func)
        except ValueError:
            sig = None
        else:
            params = dict(sig.parameters)

            # check argument names if **kwargs is not in the function
            if not any(
                param.kind
                in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
                for param in params.values()
            ):
                new_params: dict[str, inspect.Parameter] = {}
                for name in kwargs:
                    if name not in params:
                        raise TypeError(
                            f"{func} does not accept keyword argument {name!r}."
                        )

                    new_params[name] = params[name].replace(default=kwargs[name])
                params.update(new_params)
                sig = sig.replace(parameters=list(params.values()))

        self: Self = partial.__new__(cls, func, **kwargs)
        self.desc = desc
        self.keys = keys
        wraps(func)(self)
        self.__signature__ = sig
        return self

    @property
    def desc(self) -> str:
        """Description of function."""
        return self._desc

    @desc.setter
    def desc(self, val) -> None:
        if val is not None:
            if not isinstance(val, str):
                raise TypeError("description must be string.")
            desc = val
        else:
            from docstring_parser import parse

            _doc = getattr(self.func, "__doc__", "")
            if _doc == "":
                return _doc
            doc = parse(_doc)
            desc = doc.short_description

        self._desc = desc
        return None

    @property
    def keys(self) -> QtKeys:
        """Key bound to the callback."""
        return self._keys

    @keys.setter
    def keys(self, val):
        # TODO: normalize
        self._keys = val

    def __repr__(self) -> str:
        cls_str = type(self).__name__
        fn_str = getattr(self.func, "__name__", repr(self.func))
        sig_str = str(self.__signature__)  # things like (x: int)
        return f"{cls_str}<{fn_str}{sig_str}>"

    def __get__(self, obj, objtype=None) -> BoundCallback:
        if obj is None:
            return self
        if isinstance(self.func, FunctionType):
            method = self.func.__get__(obj, objtype)
        else:
            method = partial(self.func, obj)
        return BoundCallback(
            method,
            desc=self.desc,
            keys=self.keys,
            kwargs=self.keywords,
        )



import sys

# Monospace font
if sys.platform == "win32":
    MonospaceFontFamily = "Consolas"
elif sys.platform == "darwin":
    MonospaceFontFamily = "Menlo"
else:
    MonospaceFontFamily = "Monospace"



from __future__ import annotations
from typing import TYPE_CHECKING
from tabulous.exceptions import SelectionRangeError
from . import _utils, _dialogs
from tabulous.widgets._source import Source

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase

_OPACITY_CONFIG = {
    "min": 0,
    "max": 1,
    "step": 0.01,
    "value": 0.8,
    "label": "opacity",
    "widget_type": "FloatSlider",
}

_BRIGHTNESS_CONFIG = {
    "min": -1,
    "max": 1,
    "step": 0.01,
    "value": 0.0,
    "label": "opacity",
    "widget_type": "FloatSlider",
}


def set_text_colormap(viewer: TableViewerBase) -> None:
    """Set text colormap to a column"""
    from tabulous._colormap import exec_colormap_dialog

    table, column_name = _utils.get_table_and_column_name(viewer)
    if cmap := exec_colormap_dialog(
        table.native._get_sub_frame(column_name),
        table.native,
    ):
        table.text_color.set(column_name, cmap, infer_parser=False)
    return None


def reset_text_colormap(viewer: TableViewerBase) -> None:
    """Reset text colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    del table.text_color[column_name]


def set_text_colormap_opacity(viewer: TableViewerBase) -> None:
    """Set opacity to the text colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    if val := _dialogs.get_value(x=_OPACITY_CONFIG, parent=viewer.native):
        table.text_color.set_opacity(column_name, val)


def invert_text_colormap(viewer: TableViewerBase) -> None:
    """Invert text colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    table.text_color.invert(column_name)


def adjust_brightness_text_colormap(viewer: TableViewerBase) -> None:
    """Adjust brightness of the text colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    if val := _dialogs.get_value(x=_BRIGHTNESS_CONFIG, parent=viewer.native):
        table.text_color.adjust_brightness(column_name, val)


def set_background_colormap(viewer: TableViewerBase) -> None:
    """Set background colormap to a column"""
    from tabulous._colormap import exec_colormap_dialog

    table, column_name = _utils.get_table_and_column_name(viewer)
    if cmap := exec_colormap_dialog(
        table.native._get_sub_frame(column_name),
        table.native,
    ):
        table.background_color.set(column_name, cmap, infer_parser=False)
    return None


def reset_background_colormap(viewer: TableViewerBase) -> None:
    """Reset background colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    del table.background_color[column_name]


def set_background_colormap_opacity(viewer: TableViewerBase) -> None:
    """Set opacity to the background colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    if val := _dialogs.get_value(x=_OPACITY_CONFIG, parent=viewer.native):
        table.background_color.set_opacity(column_name, val)


def invert_background_colormap(viewer: TableViewerBase) -> None:
    """Invert background colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    table.background_color.invert(column_name)


def adjust_brightness_background_colormap(viewer: TableViewerBase) -> None:
    """Adjust brightness of the background colormap"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    if val := _dialogs.get_value(x=_BRIGHTNESS_CONFIG, parent=viewer.native):
        table.background_color.adjust_brightness(column_name, val)


def set_text_formatter(viewer: TableViewerBase) -> None:
    """Set text formatter"""
    from tabulous._text_formatter import exec_formatter_dialog

    table, column_name = _utils.get_table_and_column_name(viewer)

    if fmt := exec_formatter_dialog(
        table.native._get_sub_frame(column_name),
        table.native,
    ):
        table.formatter.set(column_name, fmt)
    return None


def reset_text_formatter(viewer: TableViewerBase) -> None:
    """Reset text formatter"""
    table, column_name = _utils.get_table_and_column_name(viewer)
    del table.formatter[column_name]


def run_groupby(viewer: TableViewerBase) -> None:
    """Group table by its columns (pd.groupby)"""
    table = _utils.get_mutable_table(viewer)

    cols = _utils.get_selected_columns(viewer)
    if len(cols) == 0:
        raise SelectionRangeError("No columns selected")
    colnames = [table.columns[c] for c in cols]
    out = table.data.groupby(by=colnames)
    table_out = viewer.add_groupby(out, name=f"{table.name}-groupby")
    table_out._source = Source.from_table(table)


def run_cut(viewer: TableViewerBase):
    """Cut a table column into bins (pd.cut)"""
    table = _utils.get_mutable_table(viewer)
    try:
        idx = _utils.get_selected_column(viewer)
        colname = table.columns[idx]
    except ValueError:
        colname = table.columns[0]
    ds = _dialogs.cut(
        df={"bind": table.data},
        column={"choices": table.columns, "value": colname},
        parent=viewer.native,
    )
    table.assign({f"{colname}_cut": ds})



from __future__ import annotations

from abc import ABC, abstractproperty
from functools import partial
from pathlib import Path
from types import MappingProxyType
import warnings
import weakref
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Generic, Union, TypeVar
from psygnal import Signal, SignalGroup
from superqt.utils import thread_worker

from ._table import Table, SpreadSheet, GroupBy, TableDisplay
from ._tablelist import TableList
from ._sample import open_sample
from ._component import Toolbar, Console, CommandPalette
from ._source import Source
from . import _doc

from tabulous import _utils, _io
from tabulous.types import SelectionType, TabPosition, _TableLike, _SingleSelection
from tabulous.exceptions import UnreachableError
from tabulous.widgets._keymap_abc import SupportKeyMap

if TYPE_CHECKING:
    from tabulous.widgets._table import TableBase
    from tabulous._qt import QMainWindow, QMainWidget
    from tabulous._qt._dockwidget import QtDockWidget
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous._qt._mainwindow._namespace import Namespace
    from qtpy.QtWidgets import QWidget
    from magicgui.widgets import Widget
    import numpy as np
    import pandas as pd

PathLike = Union[str, Path, bytes]
_T = TypeVar("_T")


class TableType(Enum):
    table = "table"
    spreadsheet = "spreadsheet"


class TableViewerSignal(SignalGroup):
    """Signal group for table viewer."""

    current_index = Signal(int)


class _AbstractViewer(ABC):
    @abstractproperty
    def current_table(self) -> TableBase | None:
        """Return the currently visible table."""

    @abstractproperty
    def current_index(self) -> int | None:
        """Return the index of currently visible table."""

    @abstractproperty
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""

    @property
    def data(self) -> pd.DataFrame | None:
        """The data of the current table if exists."""
        table = self.current_table
        if table is None:
            return None
        return table.data

    @property
    def config(self) -> MappingProxyType:
        """Return the config info."""
        return _utils.get_config().as_immutable()

    def reset_choices(self, *_):
        pass

    def copy_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
        *,
        headers: bool = False,
        sep: str = "\t",
    ) -> None:
        """Copy selected cells to clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.copyToClipboard(headers=headers, sep=sep)

    def paste_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
    ) -> None:
        """Paste from clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.pasteFromClipBoard()


class IPyProperty(Generic[_T]):
    """
    Essentially identical to the getter-only property.

    IPython runtime auto-completion checks if the evaluation has any side effects.
    If an attribute is a property, it immediately skips the auto-completion check.
    """

    def __init__(self, fget: Callable[[Any], _T]):
        self.fget = fget

    def __get__(self, instance, owner=None) -> _T:
        if instance is None:
            raise AttributeError("Can only be accessed via an instance.")
        return self.fget(instance)


class TableViewerBase(_AbstractViewer, SupportKeyMap):
    """The base class of a table viewer widget."""

    events: TableViewerSignal
    _qwidget_class: type[_QtMainWidgetBase]

    toolbar = Toolbar()
    console = Console()
    command_palette = CommandPalette()

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        show: bool = True,
    ):
        from tabulous._qt import get_app
        from tabulous._utils import get_post_initializers

        app = get_app()  # noqa: F841
        self._qwidget = self._qwidget_class(tab_position=tab_position)
        self._qwidget._table_viewer = self
        self._tablist = TableList(parent=self)
        self._link_events()

        self.events = TableViewerSignal()
        self._table_initializer = None
        _init = get_post_initializers()
        if _init is not None:
            viewer_initializer, self._table_initializer = _init
            viewer_initializer.initialize_viewer(self)

        if show:
            self.show(run=False)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} widget at {hex(id(self))}>"

    @IPyProperty
    def current_table(self) -> TableBase | None:
        """Return the currently visible table."""
        if len(self.tables) > 0:
            return self.tables[self.current_index]
        return None

    @property
    def current_index(self) -> int | None:
        """Return the index of currently visible table."""
        stack = self._qwidget._tablestack
        if stack.isEmpty():
            return None
        return stack.currentIndex()

    @current_index.setter
    def current_index(self, index: int | str):
        if isinstance(index, str):
            index = self.tables.index(index)
        elif index < 0:
            index += len(self.tables)
        return self._qwidget._tablestack.setCurrentIndex(index)

    @IPyProperty
    def tables(self) -> TableList:
        """Return the table list object."""
        return self._tablist

    @property
    def native(self) -> _QtMainWidgetBase:
        """Return the native widget."""
        return self._qwidget

    @property
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""
        return self._qwidget._namespace

    @property
    def history_manager(self):
        """The file I/O history manager."""
        return self._qwidget._hist_mgr

    @property
    def theme(self) -> str:
        """Get current widget theme name."""
        return self._qwidget._style_theme

    @theme.setter
    def theme(self, theme: str):
        self._qwidget.applyTheme(theme)

    def show(self, *, run: bool = True) -> None:
        """Show the widget."""
        self._qwidget.show()
        if run:
            from tabulous._qt._app import run_app

            run_app()
        return None

    def screenshot(self) -> np.ndarray:
        """Get screenshot of the widget."""
        return self._qwidget.screenshot()

    def save_screenshot(self, path: str):
        """Save screenshot of the widget."""
        from PIL import Image

        arr = self.screenshot()
        Image.fromarray(arr).save(path)
        return None

    @_doc.update_doc
    def add_table(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = False,
        copy: bool = True,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> Table:
        """
        Add data as a table.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        {name}{editable}{copy}{metadata}{update}

        Returns
        -------
        Table
            The added table object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = Table(data, name=name, editable=editable, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_spreadsheet(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = True,
        copy: bool = True,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
        dtyped: bool = False,
    ) -> SpreadSheet:
        """
        Add data as a spreadsheet.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        {name}{editable}{copy}{metadata}{update}
        dtyped : bool, default is False
            If True, dtypes of the dataframe columns will be saved in the spreadsheet.
            Typed spreadsheet is safer for data recovery but less flexible for editing.

        Returns
        -------
        SpreadSheet
            The added spreadsheet object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = SpreadSheet(data, name=name, editable=editable, metadata=metadata)
        if dtyped:
            for icol, dtype in enumerate(data.dtypes):
                col = table.columns[icol]
                table.dtypes.set(col, dtype)
            table.undo_manager.clear()
        return self.add_layer(table, update=update)

    def add_groupby(
        self,
        data,
        *,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> GroupBy:
        """
        Add a groupby object.

        Parameters
        ----------
        data : DataFrameGroupBy like object
            The groupby object to add.
        {name}{metadata}{update}

        Returns
        -------
        GroupBy
            A groupby table.
        """
        table = GroupBy(data, name=name, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_loader(
        self,
        loader: Callable[[], _TableLike],
        *,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        update: bool = False,
    ) -> TableDisplay:
        """
        Add a data frame loader function and continuously update the table.

        Parameters
        ----------
        loader : callable
            The loader function.
        {name}{metadata}{update}

        Returns
        -------
        TableDisplay
            A table display object.
        """
        table = TableDisplay(loader, name=name, metadata=metadata)
        return self.add_layer(table, update=update)

    def add_layer(self, input: TableBase, *, update: bool = False):
        """Add any table object to the viewer."""
        if table := self.current_table:
            table_has_focus = table.native._qtable_view.hasFocus()
        else:
            table_has_focus = False
        if (
            update
            and (table := self.tables.get(input.name, None))
            and type(table) is type(input)
        ):
            table.data = input.data
            return table
        self.tables.append(input)
        self.current_index = -1  # activate the last table

        if self._table_initializer is not None:
            self._table_initializer.initialize_table(input)
        if table_has_focus:
            self._qwidget.setCellFocus()
        return input

    def open(self, path: PathLike, *, type: TableType | str = TableType.table) -> None:
        """
        Read a table data and add to the viewer.

        Parameters
        ----------
        path : path like
            File path.
        """
        path = Path(path)
        type = TableType(type)
        if type is TableType.table:
            fopen = self.add_table
        elif type is TableType.spreadsheet:
            fopen = self.add_spreadsheet
        else:
            raise UnreachableError(type)

        out = _io.open_file(path)
        if isinstance(out, dict):
            for sheet_name, df in out.items():
                table = fopen(df, name=sheet_name)
                table._source = Source(path)
        else:
            table = fopen(out, name=path.stem)
            table._source = Source(path)

        _utils.dump_file_open_path(path)
        return None

    def save(self, path: PathLike) -> None:
        """Save current table."""
        warnings.warn(
            "viewer.save() is deprecated. Use table.save() instead.",
            DeprecationWarning,
        )
        _io.save_file(path, self.current_table.data)
        return None

    def save_all(self, path: PathLike) -> None:
        """Save all tables."""
        path = Path(path)
        if path.is_dir():
            paths = [path / f"{table.name}.csv" for table in self.tables]
        elif path.name.count("*") == 1:
            paths = [path.replace("*", table.name) for table in self.tables]
        elif path.suffix in (".xlsx", ".xls"):
            import pandas as pd
            from tabulous._pd_index import is_ranged

            with pd.ExcelWriter(path, mode="w") as writer:
                for table in self.tables:
                    df = table.data
                    df.to_excel(
                        writer,
                        sheet_name=table.name,
                        index=not is_ranged(df.index),
                        header=not is_ranged(df.columns),
                    )
            return None
        else:
            raise ValueError("Invalid path.")

        for table, fp in zip(self.tables, paths):
            _io.save_file(fp, table.data)
        return None

    def open_sample(
        self,
        sample_name: str,
        *,
        plugin: str = "seaborn",
        type: TableType | str = TableType.table,
        asynchronous: bool = False,
    ) -> Table:
        """Open a sample table."""
        type = TableType(type)
        if type is TableType.table:
            fopen = partial(self.add_table, name=sample_name)
        elif type is TableType.spreadsheet:
            fopen = partial(self.add_spreadsheet, name=sample_name)
        else:
            raise UnreachableError(type)
        if asynchronous:
            worker = thread_worker(open_sample)(sample_name, plugin)
            worker.returned.connect(fopen)
            self.native._tablestack._info_stack.addWorker(
                worker, f"Fetching {sample_name!r} data."
            )
            worker.start()
        else:
            df = open_sample(sample_name, plugin)
            return fopen(df)

    def copy_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
        *,
        headers: bool = False,
        sep: str = "\t",
    ) -> None:
        """Copy selected cells to clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.copyToClipboard(headers=headers, sep=sep)

    def paste_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
    ) -> None:
        """Paste from clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.pasteFromClipBoard()

    def close(self):
        """Close the viewer."""
        return self._qwidget.close()

    @property
    def status(self) -> str:
        """Return the statup tip"""
        return self._status

    @status.setter
    def status(self, tip: str) -> None:
        """Set the status tip"""
        self._status = tip
        statusbar = self._qwidget.statusBar()
        return statusbar.showMessage(tip)

    def resize(self, width: int, height: int):
        """Resize the table viewer."""
        warnings.warn(
            "viewer.resize() is deprecated. Use `viewer.size` instead.",
            DeprecationWarning,
        )
        self.size = width, height

    @property
    def size(self) -> tuple[int, int]:
        """Return the size of the table viewer."""
        return self._qwidget.size().width(), self._qwidget.size().height()

    @size.setter
    def size(self, size: tuple[int, int]):
        """Set the size of the table viewer."""
        w, h = size
        return self._qwidget.resize(int(w), int(h))

    def _link_events(self):
        _tablist = self.tables
        _qtablist = self._qwidget._tablestack

        @_tablist.events.inserted.connect
        def _insert_qtable(i: int):
            table = _tablist[i]
            _qtablist.addTable(table._qwidget, table.name)

        @_tablist.events.removed.connect
        def _remove_qtable(index: int, table: TableBase):
            _qtablist.blockSignals(True)
            try:
                _qtablist.takeTable(index)
            finally:
                _qtablist.blockSignals(False)

        @_tablist.events.moved.connect
        def _move_qtable(src: int, dst: int):
            # Evented list emits (0, 1) when move(0, 2) is called (v0.3.5 now).
            _qtablist.blockSignals(True)
            try:
                _qtablist.moveTable(src, dst)
            finally:
                _qtablist.blockSignals(False)

        @_tablist.events.renamed.connect
        def _rename_qtable(index: int, name: str):
            with _tablist.events.blocked():
                _qtablist.renameTable(index, name)

        @_qtablist.itemMoved.connect
        def _move_pytable(src: int, dst: int):
            """Move evented list when list is moved in GUI."""
            if src < dst:
                dst += 1
            with self.tables.events.blocked():
                self.tables.move(src, dst)

        @_qtablist.tableRenamed.connect
        def _rename_pytable(index: int, name: str):
            self.tables.rename(index, name)

        @_qtablist.tableRemoved.connect
        def _remove_pytable(index: int):
            with self.tables.events.blocked():
                del self.tables[index]

        @_qtablist.tablePassed.connect
        def _pass_pytable(src, index: int, dst):
            src_ = _find_parent_table(src)
            dst_ = _find_parent_table(dst)
            dst_.tables.append(src_.tables.pop(index))

        _qtablist.itemDropped.connect(self.open)

        # reset choices when something changed in python table list
        _tablist.events.inserted.connect(self.reset_choices)
        _tablist.events.removed.connect(self.reset_choices)
        _tablist.events.moved.connect(self.reset_choices)
        _tablist.events.changed.connect(self.reset_choices)
        _tablist.events.renamed.connect(self.reset_choices)


class TableViewerWidget(TableViewerBase):
    """The non-main table viewer widget."""

    events: TableViewerSignal
    _qwidget: QMainWidget

    @property
    def _qwidget_class(self) -> QMainWidget:
        from tabulous._qt import QMainWidget

        return QMainWidget

    def add_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
    ):
        """Add a widget to the viewer as a child floating widget."""
        backend_widget, name = _normalize_widget(widget, name)
        backend_widget.setParent(self._qwidget, backend_widget.windowFlags())
        return backend_widget


class TableViewer(TableViewerBase):
    """The main table viewer widget."""

    events: TableViewerSignal
    _dock_widgets: weakref.WeakValueDictionary[str, QtDockWidget]
    _qwidget: QMainWindow

    @property
    def _qwidget_class(self) -> QMainWindow:
        from tabulous._qt import QMainWindow

        return QMainWindow

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        show: bool = True,
    ):
        self._dock_widgets = weakref.WeakValueDictionary()
        self._status = ""
        super().__init__(
            tab_position=tab_position,
            show=show,
        )

    @property
    def config(self) -> MappingProxyType:
        """Return the config info."""
        from tabulous._utils import get_config

        return get_config().as_immutable()

    def add_dock_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] | None = None,
    ):
        """
        Add a Qt or magic widget as a dock widget of the table viewer.

        Parameters
        ----------
        widget : magicgui.widgets.Widget or QWidget
            Widget to add to the table viewer.
        name : str, optional
            Name of the widget.
        area : str, default is "right"
            The initial area of the dock widget.
        allowed_areas : list of str, optional
            The allowed areas of the dock widget.
        """
        backend_widget, name = _normalize_widget(widget, name)

        dock = self._qwidget.addDockWidget(
            backend_widget, name=name, area=area, allowed_areas=allowed_areas
        )
        dock.setSourceObject(widget)
        self._dock_widgets[name] = dock
        dock.show()
        return dock

    def remove_dock_widget(self, name_or_widget: str | Widget | QWidget):
        """Remove dock widget from the table viewer."""
        if isinstance(name_or_widget, str):
            name = name_or_widget
            dock = self._dock_widgets[name_or_widget]
        else:
            for k, v in self._dock_widgets.items():
                if v is name_or_widget:
                    name = k
                    dock = v
                    break
            else:
                raise ValueError(f"Widget {name_or_widget!r} not found.")
        self._qwidget.removeDockWidget(dock)
        self._dock_widgets.pop(name)
        return None

    def reset_choices(self, *_):
        """Reset all the magicgui combo boxes."""
        for dock in self._dock_widgets.values():
            widget = dock.widget
            if hasattr(widget, "reset_choices"):
                widget.reset_choices()


class DummyViewer(_AbstractViewer):
    """
    The dummy object that emulates the table viewer.

    This is used when a table is used without a viewer, while make many commands
    still available from the table.
    """

    _namespace: Namespace | None = None

    def __init__(self, table: TableBase):
        self._table = table

    @property
    def current_table(self) -> TableBase:
        return self._table

    @property
    def current_index(self) -> int:
        return 0

    @property
    def cell_namespace(self) -> Namespace:
        """Return the namespace of the cell editor."""
        cls = self.__class__
        if cls._namespace is None:
            from tabulous._qt._mainwindow._namespace import Namespace
            from tabulous._utils import load_cell_namespace

            cls._namespace = Namespace()

            # update with user namespace
            cls._namespace.update_safely(load_cell_namespace())
        return cls._namespace


def _normalize_widget(widget: Widget | QWidget, name: str) -> tuple[QWidget, str]:
    if hasattr(widget, "native"):
        backend_widget = widget.native
        if not name:
            name = widget.name
    else:
        backend_widget = widget
        if not name:
            name = backend_widget.objectName()

    return backend_widget, name


def _copy_dataframe(data) -> pd.DataFrame:
    import pandas as pd

    if isinstance(data, pd.DataFrame):
        return data.copy()
    else:
        return Table._normalize_data(data)


def _find_parent_table(qwidget: _QtMainWidgetBase) -> TableViewerBase:
    x = qwidget
    while (parent := x.parent()) is not None:
        x = parent
        if hasattr(x, "_table_viewer"):
            return x._table_viewer
    raise RuntimeError



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt, Signal, Property

try:
    from magicgui.widgets.bases import ButtonWidget, CategoricalWidget
except ImportError:
    from magicgui.widgets._bases import ButtonWidget, CategoricalWidget
from magicgui.backends._qtpy.widgets import (
    QBaseButtonWidget,
    RadioButtons as RadioButtonsBase,
    QBaseValueWidget,
)
from magicgui.widgets import RadioButtons, Select


class QToggleSwitch(QtW.QWidget):
    """
    A iPhone style toggle switch.
    See https://stackoverflow.com/questions/14780517/toggle-switch-in-qt

    Properties
    ----------
    - onColor: QtGui.QColor
    - offColor: QtGui.QColor
    - handleColor: QtGui.QColor
    """

    toggled = Signal(bool)

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Expanding
        )
        self._height = 16
        self._on_color = QtGui.QColor("#4D79C7")
        self._off_color = QtGui.QColor("#909090")
        self._handle_color = QtGui.QColor("#d5d5d5")
        self.offset = self._height / 2
        self._checked = False
        self._margin = 3
        self._anim = QtCore.QPropertyAnimation(self, b"offset", self)

        self.setFixedSize(38, 22)
        self.toggled.connect(self._set_checked)

    @Property(QtGui.QColor)
    def onColor(self):
        return self._on_color

    @onColor.setter
    def onColor(self, brsh: QtGui.QBrush):
        self._on_color = brsh

    @Property(QtGui.QColor)
    def offColor(self):
        return self._off_color

    @offColor.setter
    def offColor(self, brsh: QtGui.QBrush):
        self._off_color = brsh

    @Property(float)
    def offset(self):
        return self._x

    @offset.setter
    def offset(self, o: float):
        self._x = o
        self.update()

    @Property(QtGui.QColor)
    def handleColor(self):
        return self._handle_color

    @handleColor.setter
    def handleColor(self, color: QtGui.QColor):
        self._handle_color = color

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(
            2 * (self._height + self._margin), self._height + 2 * self._margin
        )

    def paintEvent(self, e):
        p = QtGui.QPainter(self)
        p.setPen(Qt.PenStyle.NoPen)
        _y = self._height / 2
        rrect = QtCore.QRect(
            self._margin,
            self._margin,
            self.width() - 2 * self._margin,
            self.height() - 2 * self._margin,
        )
        if self.isEnabled():
            p.setBrush(self._on_color if self._checked else self._off_color)
            p.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
            p.setOpacity(0.8)
        else:
            p.setBrush(self._off_color)
            p.setOpacity(0.6)
        p.drawRoundedRect(rrect, _y, _y)
        p.setBrush(self._handle_color)
        p.setOpacity(1.0)
        p.drawEllipse(QtCore.QRectF(self.offset - _y, 0, self.height(), self.height()))

    def mouseReleaseEvent(self, e: QtGui.QMouseEvent):
        if e.button() & Qt.MouseButton.LeftButton:
            self.toggled.emit(not self.isChecked())
        return super().mouseReleaseEvent(e)

    def enterEvent(self, e):
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        return super().enterEvent(e)

    def toggle(self):
        return self.setChecked(not self.isChecked())

    def isChecked(self) -> bool:
        return self._checked

    def setChecked(self, val: bool):
        self._set_checked(val)
        self.toggled.emit(val)

    def _set_checked(self, val: bool):
        start = self.positionForValue(self._checked)
        end = self.positionForValue(val)

        # Do not re-animate if the value is the same
        if self._checked == val:
            return
        self._checked = val
        self._anim.setStartValue(start)
        self._anim.setEndValue(end)
        self._anim.setDuration(120)
        self._anim.start()

    def positionForValue(self, val: bool) -> int:
        if val:
            return int(self.width() - self._height)
        else:
            return self._height // 2


class QLabeledToggleSwitch(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._switch = QToggleSwitch(self)
        self._text = QtW.QLabel(self)
        layout.addWidget(self._switch)
        layout.addWidget(self._text)
        self.setLayout(layout)

    @property
    def toggled(self):
        return self._switch.toggled

    def enterEvent(self, e):
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        return super().enterEvent(e)

    def isChecked(self) -> bool:
        return self._switch.isChecked()

    def setChecked(self, val: bool):
        self._switch.setChecked(val)

    def isDown(self) -> bool:
        return self.isChecked()

    def setDown(self, a0: bool) -> None:
        return self.setChecked(a0)

    def text(self) -> str:
        return self._text.text()

    def setText(self, text: str):
        self._text.setText(text)

    def click(self):
        self.toggle()

    def toggle(self):
        self.setChecked(not self.isChecked())


class QButtonGroup(QtCore.QObject):
    """A button group object that mimics the Qt API."""

    buttonToggled = Signal(QLabeledToggleSwitch, bool)

    def __init__(self) -> None:
        super().__init__()
        self._list: list[QLabeledToggleSwitch] = []
        self._exclusive = True

    def buttons(self) -> list[QLabeledToggleSwitch]:
        """Return the list of buttons in the group."""
        # NOTE: Must return a copy of the list because this function will be used as
        # an iterator.
        return self._list.copy()

    def checkedButton(self) -> QLabeledToggleSwitch:
        for btn in self._list:
            if btn.isChecked():
                return btn
        return None

    def checkedButtons(self) -> list[QLabeledToggleSwitch]:
        return [btn for btn in self._list if btn.isChecked()]

    def addButton(self, btn: QLabeledToggleSwitch):
        self._list.append(btn)
        btn._switch.toggled.connect(
            lambda checked: self._on_button_toggled(btn, checked)
        )

    def removeButton(self, btn: QLabeledToggleSwitch):
        self._list.remove(btn)
        btn._switch.toggled.disconnect()

    def exclusive(self) -> bool:
        return self._exclusive

    def setExclusive(self, val: bool):
        self._exclusive = val

    def _on_button_toggled(self, btn: QLabeledToggleSwitch, checked: bool):
        if self.exclusive():
            if checked:
                for b in self._list:
                    if b is not btn:
                        b._switch._set_checked(False)
            else:
                btn._switch._set_checked(True)
        self.buttonToggled.emit(btn, checked)


class QToggleSwitchesBase(RadioButtonsBase):
    _qwidget: QtW.QGroupBox

    def __init__(self, **kwargs):
        QBaseValueWidget.__init__(self, QtW.QGroupBox, "", "", "", **kwargs)
        self._btn_group = QButtonGroup()
        self._qwidget.setLayout(QtW.QVBoxLayout())
        self._qwidget.layout().setAlignment(Qt.AlignmentFlag.AlignTop)
        self._btn_group.buttonToggled.connect(self._emit_data)

    def _emit_data(self, btn, checked):
        if checked or not self._btn_group.exclusive():
            self._event_filter.valueChanged.emit(self._mgui_get_value())

    def _add_button(self, label: str, data=None):
        btn = QLabeledToggleSwitch(self._qwidget)
        btn.setText(label)
        btn._data = data
        self._btn_group.addButton(btn)
        self._qwidget.layout().addWidget(btn)

    def _mgui_get_value(self):
        if self._btn_group.exclusive():
            btn = self._btn_group.checkedButton()
            return btn._data if btn else None
        else:
            return [btn._data for btn in self._btn_group.checkedButtons()]

    def _mgui_set_value(self, value) -> None:
        if self._btn_group.exclusive():
            for btn in self._btn_group.buttons():
                if btn._data == value:
                    btn._switch._set_checked(True)
                    break  # exclusive
        else:
            for btn in self._btn_group.buttons():
                btn._switch._set_checked(btn._data in value)


class ToggleSwitch(ButtonWidget):
    def __init__(self, **kwargs):
        super().__init__(
            widget_type=QBaseButtonWidget,
            backend_kwargs={"qwidg": QLabeledToggleSwitch},
            **kwargs,
        )
        self.native: QLabeledToggleSwitch


class ToggleSwitches(RadioButtons):  # type: ignore
    """An radio buttons using toggle switches."""

    def __init__(self, choices=(), orientation="vertical", **kwargs):
        kwargs["widget_type"] = QToggleSwitchesBase
        CategoricalWidget.__init__(self, choices=choices, **kwargs)
        self.orientation = orientation


class ToggleSwitchSelect(Select):
    def __init__(self, choices=(), orientation="vertical", **kwargs):
        kwargs["widget_type"] = QToggleSwitchesBase
        CategoricalWidget.__init__(self, choices=choices, **kwargs)
        self._widget._btn_group.setExclusive(False)
        self.orientation = orientation



from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtWidgets import QFileSystemModel
from qtpy.QtCore import Qt

from magicgui.widgets import FileEdit

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


class QFileTreeWidget(QtW.QTreeView):
    """A file tree widget for tabulous"""

    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        self.setRootIsDecorated(True)
        self.setAnimated(True)
        self.setIndentation(20)
        self.setTextElideMode(Qt.TextElideMode.ElideNone)
        self.setDropIndicatorShown(True)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_context_menu)
        self.setUniformRowHeights(True)
        self.setAlternatingRowColors(False)
        self.setExpandsOnDoubleClick(True)
        self.setModel(QFileSystemModel())
        for i in range(1, self.model().columnCount()):
            self.hideColumn(i)
        self.doubleClicked.connect(self._open_at_index)

    def model(self) -> QFileSystemModel:
        return super().model()

    def setRoot(self, path: str):
        path = str(path)
        index = self.model().setRootPath(path)
        self.setHeaderHidden(True)
        self.setRootIndex(index)

    def _on_context_menu(self, point: QtCore.QPoint):
        menu = QtW.QMenu(self)
        index = self.indexAt(point)
        menu.addAction(
            "Open as table", lambda: self._open_at_index(index, type="table")
        )
        menu.addAction(
            "Open as spreadsheet",
            lambda: self._open_at_index(index, type="spreadsheet"),
        )
        menu.exec_(self.viewport().mapToGlobal(point))

    def _find_viewer(self) -> TableViewerBase | None:
        parent = self.parentWidget()
        while parent is not None:
            parent = parent.parentWidget()
            if hasattr(parent, "_table_viewer"):
                return parent._table_viewer
        return None

    def _open_at_index(self, index: QtCore.QModelIndex, type: str = "table"):
        path = self.model().filePath(index)
        if not self.model().isDir(index):
            if viewer := self._find_viewer():
                path = self.model().filePath(index)
                viewer.open(path, type=type)


class QFileExplorer(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        self._layout = QtW.QVBoxLayout(self)
        self._file_edit = FileEdit(mode="d")
        self._file_edit.line_edit.tooltip = "Root directory"
        self._file_edit.choose_btn.text = "..."
        self._file_tree = QFileTreeWidget(self)
        curpath = QtCore.QDir.currentPath()
        self._file_tree.setRoot(curpath)
        self._file_edit.value = Path(curpath)
        self._layout.addWidget(self._file_edit.native)
        self._layout.addWidget(self._file_tree)
        self._file_edit.changed.connect(self._update_root)
        self._update_root(curpath)

    def _update_root(self, path: str):
        if Path(path).exists():
            self._file_tree.setRoot(str(path))



from __future__ import annotations

from contextlib import contextmanager
from typing import (
    Any,
    Iterable,
    Iterator,
    MutableMapping,
    NamedTuple,
    TYPE_CHECKING,
    TypeVar,
)
from psygnal import Signal
import logging
import weakref
from tabulous._range import TableAnchorBase
from tabulous._psygnal import InCellRangedSlot

if TYPE_CHECKING:
    from tabulous.widgets import TableBase


class Index(NamedTuple):
    row: int
    column: int


_V = TypeVar("_V")

logger = logging.getLogger(__name__)


class TableMapping(MutableMapping[Index, _V], TableAnchorBase):
    set = Signal(Index, object)
    deleted = Signal(object)

    def __init__(self) -> None:
        self._dict: dict[Index, _V] = {}

    def __getitem__(self, key: Index) -> _V:
        return self._dict[key]

    def __setitem__(self, key: tuple[int, int], value: _V) -> None:
        logger.debug(f"Setting TableMapping item at {key}")
        index = Index(*key)
        if index in self._dict:
            del self[index]
        self._dict[index] = value
        self.set.emit(key, value)

    def __delitem__(self, key: Index) -> None:
        logger.debug(f"Deleting TableMapping item at {key}")
        value = self._dict.pop(key)
        self.deleted.emit(value)

    def __iter__(self) -> Iterator[Index]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def insert_rows(self, row: int, count: int):
        """Insert rows and update indices."""
        new_dict = {}
        for idx in list(self._dict.keys()):
            if idx.row >= row:
                new_idx = Index(idx.row + count, idx.column)
                child = self._dict.pop(idx)
                new_dict[new_idx] = child
                # if isinstance(child, TableAnchorBase):
                #     child.insert_rows(row, count)

        self._dict.update(new_dict)
        return None

    def insert_columns(self, col: int, count: int):
        """Insert columns and update indices."""
        new_dict = {}
        for idx in list(self._dict.keys()):
            if idx.column >= col:
                new_idx = Index(idx.row, idx.column + count)
                child = self._dict.pop(idx)
                new_dict[new_idx] = child
                # if isinstance(child, TableAnchorBase):
                #     child.insert_columns(col, count)

        self._dict.update(new_dict)
        return None

    def remove_rows(self, row: int, count: int):
        """Remove items that are in the given row range."""
        start = row
        stop = row + count
        for idx in list(self._dict.keys()):
            if start <= idx.row < stop:
                self.pop(idx)
            elif idx.row >= stop:
                new_idx = Index(idx.row - count, idx.column)
                child = self._dict.pop(idx)
                self._dict[new_idx] = child
                # if isinstance(child, TableAnchorBase):
                #     child.remove_rows(row, count)

        return None

    def remove_columns(self, col: int, count: int):
        """Remove items that are in the given column range."""
        start = col
        stop = col + count
        for idx in list(self._dict.keys()):
            if start <= idx.column < stop:
                self.pop(idx)
            elif idx.column >= stop:
                new_idx = Index(idx.row, idx.column - count)
                child = self._dict.pop(idx)
                self._dict[new_idx] = child
                # if isinstance(child, TableAnchorBase):
                #     child.remove_columns(col, count)

        return None


class SlotRefMapping(MutableMapping[Index, InCellRangedSlot], TableAnchorBase):
    def __init__(self, table: TableBase) -> None:
        self._table_ref = weakref.ref(table)
        self._locked_pos = None

    def table(self) -> TableBase:
        if table := self._table_ref():
            return table
        raise RuntimeError("Table has been deleted")

    def __getitem__(self, source_key: Index) -> InCellRangedSlot:
        for slot in self.table().events.data.iter_slots():
            if not isinstance(slot, InCellRangedSlot):
                continue
            if slot.source_pos == source_key:
                return slot
        raise KeyError(source_key)

    def get_by_dest(self, source_key: Index, default=None) -> InCellRangedSlot:
        for slot in self.table().events.data.iter_slots():
            if not isinstance(slot, InCellRangedSlot):
                continue
            if slot.source_pos == source_key:
                return slot
            dest = slot.last_destination
            if dest is not None:
                rsl, csl = dest
                r, c = source_key
                if rsl.start <= r < rsl.stop and csl.start <= c < csl.stop:
                    return slot
        return default

    def __setitem__(self, key: Index, slot: InCellRangedSlot) -> None:
        if self._locked_pos == key:
            return
        if self.pop(key, None):
            logger.debug(f"Overwriting slot at {key}")
        self.table().events.data.connect_cell_slot(slot)
        logger.debug(f"Connecting slot at {key}")

    def __delitem__(self, source_key: Index) -> None:
        if self._locked_pos == source_key:
            return
        slot = self[source_key]
        self.table().events.data.disconnect(slot)
        logger.debug(f"Deleting slot at {source_key}")

    def _remove_multiple(self, slots: Iterable[InCellRangedSlot]):
        data = self.table().events.data
        for slot in slots:
            data.disconnect(slot)

    def __iter__(self) -> Iterator[Index]:
        for slot in self.table().events.data.iter_slots():
            if not isinstance(slot, InCellRangedSlot):
                continue
            yield Index(*slot.source_pos)

    def values(self):
        for slot in self.table().events.data.iter_slots():
            if not isinstance(slot, InCellRangedSlot):
                continue
            yield slot

    def items(self):
        for slot in self.table().events.data.iter_slots():
            if not isinstance(slot, InCellRangedSlot):
                continue
            yield Index(*slot.source_pos), slot

    def __len__(self) -> int:
        return len([i for i in self])

    @contextmanager
    def lock_pos(self, pos: Index):
        _old_pos = self._locked_pos
        self._locked_pos = pos
        try:
            yield
        finally:
            self._locked_pos = _old_pos

    def insert_rows(self, row: int, count: int):
        """Insert rows and update indices."""

    def insert_columns(self, col: int, count: int):
        """Insert columns and update indices."""

    def remove_rows(self, row: int, count: int):
        """Remove items that are in the given row range."""
        start = row
        stop = row + count
        rem = []
        for idx, slot in self.items():
            if start <= idx.column < stop:
                rem.append(slot)
        self._remove_multiple(rem)
        return None

    def remove_columns(self, col: int, count: int):
        """Remove items that are in the given column range."""
        start = col
        stop = col + count
        rem = []
        for idx, slot in self.items():
            if start <= idx.column < stop:
                rem.append(slot)
        self._remove_multiple(rem)
        return None


class DummySlotRefMapping(MutableMapping[Index, Any], TableAnchorBase):
    def table(self) -> TableBase:
        raise RuntimeError("This is dummy mapping")

    def __getitem__(self, key: Index) -> InCellRangedSlot:
        raise KeyError(key)

    def get_by_dest(self, key: Index, default=None) -> InCellRangedSlot:
        return default

    def __setitem__(self, key: Index, slot: InCellRangedSlot) -> None:
        raise KeyError(key)

    def __delitem__(self, key: Index) -> None:
        raise KeyError(key)

    def __iter__(self) -> Iterator[Index]:
        raise StopIteration

    def __len__(self) -> int:
        return 0

    @contextmanager
    def lock_pos(self, pos: Index):
        yield

    def insert_rows(self, row: int, count: int):
        """Insert rows and update indices."""

    def insert_columns(self, col: int, count: int):
        """Insert columns and update indices."""

    def remove_rows(self, row: int, count: int):
        """Remove items that are in the given row range."""

    def remove_columns(self, col: int, count: int):
        """Remove items that are in the given column range."""



from __future__ import annotations

from ._mainwidgets import QMainWindow, QMainWidget, _QtMainWidgetBase

# fmt: off

@QMainWidget._keymap.bind("Ctrl+K")
@QMainWindow._keymap.bind("Ctrl+K")
def _(self: _QtMainWidgetBase):
    """Basic sequence to trigger key combo."""
    return self.setFocus()

@QMainWidget._keymap.bind_deactivated("Ctrl+K")
@QMainWindow._keymap.bind_deactivated("Ctrl+K")
def _(self: _QtMainWidgetBase):
    return self.setCellFocus()


@QMainWindow._keymap.bind("Alt")
def _(self: _QtMainWidgetBase):
    """Move focus to toolbar."""
    self._toolbar.showTabTooltips()
    self.setFocus()


@QMainWindow._keymap.bind("Alt, H", index=0, desc="Move focus to `Home` menu tab.")
@QMainWindow._keymap.bind("Alt, E", index=1, desc="Move focus to `Edit` menu tab.")
@QMainWindow._keymap.bind("Alt, T", index=2, desc="Move focus to `Table` menu tab.")
@QMainWindow._keymap.bind("Alt, A", index=3, desc="Move focus to `Analyze` menu tab.")
@QMainWindow._keymap.bind("Alt, V", index=4, desc="Move focus to `View` menu tab.")
@QMainWindow._keymap.bind("Alt, P", index=5, desc="Move focus to `Plot` menu tab.")
def _(self: QMainWindow, index: int):
    self._toolbar.setCurrentIndex(index)
    self._toolbar.currentToolBar().showTabTooltips()


@QMainWindow._keymap.bind("Alt, H, {}")
@QMainWindow._keymap.bind("Alt, E, {}")
@QMainWindow._keymap.bind("Alt, T, {}")
@QMainWindow._keymap.bind("Alt, A, {}")
@QMainWindow._keymap.bind("Alt, V, {}")
@QMainWindow._keymap.bind("Alt, P, {}")
def _(self: QMainWindow, key: str):
    """Push a tool button at the given position."""
    try:
        index = int(key)
    except ValueError:
        return None

    self._toolbar.currentToolBar().clickButton(
        (index - 1) % 10, ignore_index_error=True
    )

# fmt: on



from __future__ import annotations
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Signal


class QClickableLabel(QtW.QLabel):
    """A label widget that behaves like a button."""

    clicked = Signal()

    def __init__(self, text: str, parent=None):
        super().__init__(parent)
        self.setFont(QtGui.QFont("Arial"))
        self.setFixedHeight(24)

        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Expanding
        )
        self.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setText(text)
        return None

    def setText(self, text: str):
        """Set the label text and resize the widget to fit the text."""
        fm = QtGui.QFontMetrics(self.font())
        width = fm.width(text)
        self.setFixedWidth(width + 18)
        return super().setText(text)

    def mouseReleaseEvent(self, ev: QtGui.QMouseEvent) -> None:
        """Emit the clicked signal when the left mouse button is released."""
        if ev.button() == Qt.MouseButton.LeftButton:
            self.clicked.emit()
        return super().mouseReleaseEvent(ev)

    def enterEvent(self, a0: QtCore.QEvent) -> None:
        """Add an underline to the text and change the cursor to a hand."""
        font = self.font()
        font.setUnderline(True)
        self.setFont(font)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.update()
        return super().enterEvent(a0)

    def leaveEvent(self, a0: QtCore.QEvent) -> None:
        """Reset the text and cursor to their original state."""
        font = self.font()
        font.setUnderline(False)
        self.setFont(font)
        self.unsetCursor()
        return super().leaveEvent(a0)
